import os
import pyneb as pn
import numpy as np
import re
from .misc import ROOT_DIR, parseAtom, roman_to_int, multi_split, int_to_roman, strExtract
from .init import ELEM_LIST, SPEC_LIST
from .physics import _predefinedDataFileDict

def atom2chianti(atom):
    elem = parseAtom(atom)[0]
    spec = int(parseAtom(atom)[1])
    return '{0}_{1}'.format(elem.lower(), spec)

class _ManageAtomicData(object):
    
    def __init__(self):
        self.log_ = pn.log_
        self.calling = '_ManageAtomicData'
        self.root_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir))
        self._predefinedDataFileDict = _predefinedDataFileDict
        self.setDataFileDict()
        self.addDataFilePath()
        self.addDataFilePath('atomic_data_fits/', inPyNeb=True)
        self.addDataFilePath('atomic_data/', inPyNeb=True)
        self.addDataFilePath('./')
        self._RecombData = {}
        self._initChianti()
        self.read_gsconf()        
        
    def includeFitsPath(self):
        self.addDataFilePath('atomic_data_fits/old_fits/', inPyNeb=True)

    def removeFitsPath(self):
        self.removeDataFilePath('atomic_data_fits/old_fits/')
        
    def getPredefinedDataFileDict(self, data_dict=None):
        """
        Retrieve all predefined dictionaries of atomic data (or a specified one if data_dict is specified)
        
        Parameters:
           - data_dict    name of the dictionary (default: None)

        """
#        Retrieve a predefined dictionary of atomic data (either the default one [data_dict unspecified]
#        or another one [identified by data_dict])
        if data_dict is None:
            return self._predefinedDataFileDict
        elif data_dict in self._predefinedDataFileDict:
            return self._predefinedDataFileDict[data_dict]
        else:
            self.log_.error('unknown DataFile dictionary {0}'.format(data_dict), calling=self.calling)
            return None


    def getAllPredefinedDict(self):
        """
        Return the labels of the available predefined dictionaries of atomic data

        """
        return pn.atomicData.getPredefinedDataFileDict().keys()

    
    def resetDataFileDict(self):
        """
        Reset the DataFileDict to the default value. This value is in defaultDict

        """
        self.setDataFileDict(self.defaultDict)


    def setDataFileDict(self, data_dict=None):
        """
        Use a dictionary for the atomic data (either predefined [string] or custom [dictionary]).
        
        Parameters:
           - data_dict    name of the dictionary 

        """
        if data_dict is None:
            self._DataFileDict = {}
            return None
        if type(data_dict) == type({}):
            for atom in data_dict:
                for data_type in data_dict[atom]:
                    self.setDataFile(data_dict[atom][data_type], atom, data_type)
            self.predefined = None
        elif data_dict in self.getPredefinedDataFileDict():
            self.setDataFileDict()
            self.setDataFileDict(self.getPredefinedDataFileDict(data_dict))
            self.predefined = data_dict
            self.log_.message('DataFileDict set to {0}.'.format(data_dict), calling=self.calling)
        else:
            self.log_.error('{0} is not a valid DataFileDict.'.format(data_dict), calling=self.calling)
 
            
    def addDataFilePath(self, fits_dir=None, inPyNeb=False):
        """
        Add a directory to the list of directories where atomic data files are searched for.
        
        Parameters:
           - fits_dir    directory
           - inPyNeb     Boolean.

        """
        if fits_dir is None:
            self._DataFilePaths = []
        else:
            try:
                if inPyNeb:
                    self._DataFilePaths.append(ROOT_DIR + '/' + fits_dir)
                else:
                    self._DataFilePaths.append(os.path.abspath(fits_dir))
            except:
                self.log_.warn('{0} could not be added to the path list', calling=self.calling)
 
 
    def removeDataFilePath(self, fits_dir=None):
        """
        Remove a directory from the list of directories where atomic data files are searched for.

        Parameters:
           - fits_dir    name of the directory to be removed

        """
        if fits_dir is None:
            pass
        else:
            try:
                self._DataFilePaths.remove(os.path.abspath(fits_dir))
            except:
                pass
            try:
                self._DataFilePaths.remove(ROOT_DIR + '/' + fits_dir)
            except:
                pass
 
                
    def getAllDataFilePaths(self):
        """
        Return the list of directories where atomic data files are searched for.

        """
        return self._DataFilePaths
 
            
    def getDirForFile(self, data_file):
        """
        Return the first directory from getDataFilePaths() where a file is found.
        If nothing is found, return None.

        Parameters:
           - data_file    name of the file

        """
        
        for dir in self._DataFilePaths:
            if data_file in os.listdir(dir):
                return dir
        if data_file.split('.')[-1] == 'chianti':
            strs = data_file.split('_')
            elem = strs[0].capitalize()
            spec = roman_to_int(strs[1])
            atom = elem + str(spec)
            strs = re.split(r'[_.]+',data_file)
            data_type = strs[2]
            if (data_type == 'atom') and (atom2chianti(atom) in self.ChiantiIONS['atom']):
                return('{}/{}/{}'.format(self.Chianti_path, elem.lower(), atom2chianti(atom)))
            if (data_type == 'coll') and (atom2chianti(atom) in self.ChiantiIONS['coll']):
                return('{}/{}/{}'.format(self.Chianti_path, elem.lower(), atom2chianti(atom)))
        else:
            return None
 
    
    def printDirForAllFiles(self):
        """
        Print the directories where all the files defined in the dictionary of 
            atomic data files are located.

        """
        for atom in self._DataFileDict:
            atom_file = self.getDataFile(atom, 'atom', warn=False)
            if atom_file is not None:
                print('{0} atom from {1} in {2}'.format(atom, atom_file , self.getDirForFile(atom_file)))
            coll_file = self.getDataFile(atom, 'coll', warn=False)
            if coll_file is not None:
                print('{0} coll from {1} in {2}'.format(atom, coll_file , self.getDirForFile(coll_file)))
            rec_file = self.getDataFile(atom, 'rec', warn=False)
            if rec_file is not None:
                print('{0} rec from {1} in {2}'.format(atom, rec_file , self.getDirForFile(rec_file)))
 
 
    def getAllAvailableFiles(self, atom=None, data_type=None):
        """
        Scan every directory in the list of paths, printing all the *atom*.fits, *atom*.dat,
            *coll*.fits and *rec*.fits files
            
        Parameters:
           - atom        atom name
           - data_type   either 'rec', 'atom', 'coll' or 'trc'

        """
        file_list = []
        if data_type is None:
            data_types = ['atom', 'coll', 'rec', 'trc']
        else:
            data_types = [data_type]
        if atom is not None:
            elem, spec = parseAtom(atom)
            elem = elem.lower()
            spec = int_to_roman(int(spec)).lower()
            atom_str = '{0}_{1}_'.format(elem, spec)
            if atom2chianti(atom) in self.ChiantiIONS['atom']:
                if 'atom' in data_types:
                    file_list.append('{0}_{1}_atom.chianti'.format(elem, spec))
            if atom2chianti(atom) in self.ChiantiIONS['coll']:
                if 'coll' in data_types:
                    file_list.append('{0}_{1}_coll.chianti'.format(elem, spec))
        else:
            atom_str = '.'
            for atom in self.ChiantiIONS['atom']:
                elem, spec = parseAtom(atom)
                elem = elem.lower()
                spec = int_to_roman(int(spec)).lower()
                if 'atom' in data_types:
                    file_list.append('{0}_{1}_atom.chianti'.format(elem, spec))
            for atom in self.ChiantiIONS['coll']:
                elem, spec = parseAtom(atom)
                elem = elem.lower()
                spec = int_to_roman(int(spec)).lower()
                if 'coll' in data_types:
                    file_list.append('{0}_{1}_coll.chianti'.format(elem, spec))
        for dir in self._DataFilePaths:
            files = os.listdir(dir)
            for file in files:
                if (('.fits' in file) or ('.func' in file) or ('.hdf5' in file) or ('.dat' in file) ) and (atom_str in file):
                    for dt in data_types:
                        if dt in file:
                            file_list.append(file)
                            
        file_list.sort()
        return file_list
    
    
    def getAllPossibleAtoms(self):
        """
        Return the list of all the possible collisional and recombination 
        atoms that can be built with the available datafiles (not all of 
        them necessarily included in the current atomic file dictionary) 
                    
        Parameters:
           - atom        atom name
           - data_type   either 'rec', 'atom' or 'coll'

        """
        atom_dict = {'atom':[], 'coll': [], 'rec': []}
        for path in self._DataFilePaths:
            files = os.listdir(path)
            for elem in ELEM_LIST:
                for spec in SPEC_LIST:
                    atom = elem + spec
                    atom_str = '{0}_{1}_'.format(elem.lower(), int_to_roman(int(spec)).lower())
                    for entry in files:
                        if (('.fits' in entry) or ('.dat' in entry) ) and (atom_str in entry):
                            if ('atom' in entry):
                                atom_dict['atom'].append(atom)
                            elif ('coll' in entry):
                                atom_dict['coll'].append(atom)
                            elif ('rec' in entry):
                                atom_dict['rec'].append(atom)
                            else:
                                pass
        coll_atom_list = list(set(atom_dict['atom']) & set(atom_dict['coll']))
        rec_atom_list = list(set(atom_dict['rec']))
        print('Data for the following collisional atoms exist:', coll_atom_list)
        print('Data for the following recombination atoms exist:', rec_atom_list)
        return [coll_atom_list, rec_atom_list] 
        
    
    def getAllAtoms(self, coll=True, rec=False):
        """
        Return a list of all the atoms (e.g. 'O3') included in the adopted dictionary 
        and for which dataFiles are available (a subset of all the ions returned by
        the getAllPossibleIons() command)

        Parameters:
            - coll     if True (default) includes the Atom objects
            - rec     if True (not the default) includes the RecAtoms

        """
        to_return = []
        for atom in self._DataFileDict:
            if coll and ('atom' in self._DataFileDict[atom]):
                to_return.append(atom)
            if rec and ('rec' in self._DataFileDict[atom]):
                to_return.append(atom)
        return to_return
            

    def getDataFile(self, atom=None, data_type=None, warn=True):
        """
        Return the name of the atomic data file associated to an atom and a type of data, 
        which is one of ('atom', 'coll', 'rec').
            
        Parameters:
            - atom        selected atom
            - data_type   type of atomic data ('atom', 'coll', 'rec', 'trc')
            - warn        warn if no associated data file is found
        
        """
        if atom is None:
            if data_type is None:
                return self._DataFileDict
            else:
                DFD = self._DataFileDict
                to_return = {}
                for at in DFD:
                    if data_type in DFD[at]:
                        to_return[at] = DFD[at][data_type]
                return to_return
        if data_type is None:
            return(self.getDataFile(atom, 'atom', warn=warn),
                   self.getDataFile(atom, 'coll', warn=warn),
                   self.getDataFile(atom, 'rec', warn=warn),
                   self.getDataFile(atom, 'trc', warn=warn))
        if data_type not in ('atom', 'coll', 'rec', 'trc'):
            if warn:
                self.log_.warn('{0} type unknown'.format(data_type), calling=self.calling)
            return None
        if atom not in self._DataFileDict:
            if warn:
                self.log_.warn('data for {0} not available'.format(atom), calling=self.calling)
            return None
        if data_type in self._DataFileDict[atom]:
            return self._DataFileDict[atom][data_type]
        else:
            if warn:
                self.log_.warn('{0} data not available for {1}'.format(data_type, atom), calling=self.calling)
            return None

        
    def getDataFullPath(self, atom, data_type=None, warn=True):
        """
        Return the full path of the atomic data file associated to an atom and a type of data, 
        which is one of ('atom', 'coll', 'rec').
        
        Parameters:
            - atom        selected atom
            - data_type   type of atomic data ('atom', 'coll', 'rec', 'trc')
            - warn        warn if no associated data file is found
        
        """
        data_file = self.getDataFile(atom, data_type=data_type, warn=warn)
        data_path = self.getDirForFile(data_file)
        if data_file is None:
            return None
        else:
            return '{0}/{1}'.format(data_path, data_file)


    def scanDirForDataFiles(self, fits_dir):
        """
        Scan a directory and associate any file named x_i_type_ref.fits to the atom Xi (i in roman), 
        type being one of ('atom', 'coll', 'rec').
        The directory must have been registered before using addDataFilePath(fits_dir)

        Parameters:
            - fits_dir        directory name

        """
        for file_ in os.listdir(fits_dir):
            names = multi_split(file_, ['_', '.'])
            elem = names[0].title()            
            if elem in ELEM_LIST:
                spec = str(roman_to_int(names[1]))
                if spec in SPEC_LIST:
                    atom = elem + spec
                    if atom not in self._DataFileDict:
                        self._DataFileDict[atom] = {}
                    if names[2] in ('atom', 'coll', 'rec', 'trc'):
                        self.setDataFile(atom, names[2], file_)
 
    def setDataFile(self, data_file=None, atom=None, data_type=None):
        """
        Associate an atomic data file to an atom and a type of data, which is one of 
            ('atom', 'coll', 'rec', 'func').

        Usage:
            pn.atomicData.setDataFile('cl_iii_atom_M83-KS86.fits')
            pn.atomicData.setDataFile('TEST.fits', 'O3', 'atom') # but the previous way 
                to name the files is preferred
            pn.atomicData.setDataFile('o_iii_atom.chianti')

        Parameters:
            - data_file    atomic data file. 
            - [atom]         selected atom
            - [data_type]    'atom', 'coll', 'rec'
            If atom and data_type not set, file format is assumed to be e.g. o_iii_coll_REF.fits

        """          
        if atom is None:
            strs = data_file.split('_')
            elem = strs[0].capitalize()
            spec = roman_to_int(strs[1])
            atom = elem + str(spec)
        else:
            elem, spec = parseAtom(atom)
        if data_type is None:
            strs = re.split(r'[_.]+',data_file)
            data_type = strs[2]
        if data_type not in ('atom', 'coll', 'rec', 'trc', 'func'):
            self.log_.error('{0} is not a valid type'.format(data_type))
            return None
        if elem not in ELEM_LIST:
            self.log_.error('{0} is not a valid element'.format(elem))
            return None
        if atom not in self._DataFileDict:
            self._DataFileDict[atom] = {}
        if self.getDirForFile(data_file) is not None:
            self._DataFileDict[atom][data_type] = data_file
            # as something changed, predefined is not valid anymore
            self.predefined = None
            if data_type == 'rec':
                if atom in self._RecombData:
                    del self._RecombData[atom]
            self.log_.message('Adding {0} {1} data for {2}'.format(data_file, data_type, atom), calling=self.calling)
        else:
            av_data = self.getAllAvailableFiles(atom, data_type)
            self.log_.error("""Unknown file {0} or corresponding File path not included in list. 
You may mean one of these files: {1}""".format(data_file, av_data),
                           calling=self.calling)
        if (atom == 'H1') and ('H1' in pn.atomicData._RecombData):
            del pn.atomicData._RecombData['H1']


    def printAllSources(self, at_set=None, predef=None):
        """
        Print bibliographic sources of the adopted data sets of a list of atoms.
        
        Usage:
        pn.atomicData.printAllSources(['O3', 'Ar4', 'S2'])
        pn.atomicData.printAllSources([O, S])
        pn.atomicData.printAllSources()

        Parameters: 
            - at_set   a list or tuple containing the atoms. If not set, 
                       print bibliographic sources for all the atoms in PyNeb's default dictionary
                       
        """        
        self.calling = 'printAllSources'
        if (type(at_set) == list) or (type(at_set) == tuple):
            at_dict = {}
            for item in at_set:
                atom = parseAtom(item)[0]
                spec = parseAtom(item)[1]
                if spec is '':
                    for ispec in SPEC_LIST:
                        try:
                            at_dict[atom+ispec] = pn.Atom(atom, ispec)
                        except:
                            pass
                else:
                    at_dict[item] = pn.Atom(atom, spec)
        elif at_set is not None:
            pn.log_.error('The argument at_set must be a list or a tuple', calling=self.calling)
            return None
        elif (at_set is None) and (predef is None):
            at_dict = pn.getAtomDict()
        elif predef in pn.atomicData.getAllPredefinedDict():
            current = pn.atomicData.predefined
            pn.atomicData.setDataFileDict(predef)
            at_dict = pn.getAtomDict()
            pn.atomicData.setDataFileDict(current)
        else: 
            pn.log_.error('The argument predef must be the label of a predefined dictionary', 
                          calling=self.calling)
        
        for item in sorted(at_dict):                                                              
            at_dict[item].printSources()
    
    def _initChianti(self):
        self.ChiantiIONS = {'atom':[], 'coll':[]}
        self.Chianti_path = None
        
        if pn.config.INSTALLED['Chianti']:
    
            self.Chianti_path = os.environ['XUVTOP']
            masterlist = '{0}/masterlist/masterlist.ions'.format(self.Chianti_path)
            try:
                with open(masterlist) as master:
                    for line in master.readlines():
                        atom = line.split()[0]
                        elem = atom.split('_')[0]
                        if pn.config.Chianti_version_main == '7':
                            coll_file = '{0}/{1}/{2}/{2}.splups'.format(self.Chianti_path, elem, atom)
                        elif pn.config.Chianti_version_main == '8':
                            coll_file = '{0}/{1}/{2}/{2}.scups'.format(self.Chianti_path, elem, atom)
                        elif pn.config.Chianti_version_main == '9':
                            coll_file = '{0}/{1}/{2}/{2}.scups'.format(self.Chianti_path, elem, atom)
                        else:
                            pn.log_.error('Unknown version of Chianti {}'.format(pn.config.Chianti_version),
                                          calling="_ManageAtomicData/_initChianti")
                        atom_file = '{0}/{1}/{2}/{2}.wgfa'.format(self.Chianti_path, elem, atom)
                        if os.path.exists(coll_file): 
                            self.ChiantiIONS['coll'].append(atom)
                        if os.path.exists(atom_file):
                            self.ChiantiIONS['atom'].append(atom)
            except:
                pn.log_.warn('File not found {}, no Chianti data available'.format(masterlist), 
                             calling='_initChianti')
    def addAllChianti(self, replace_all=False):
        """
        Adding all the Chianti file to the list of data available
        replace_all: if True, replace already available data by the Chianti ones:
            all the atoms will be Chianti (except HI, HeI and HeII)
        """
        atoms = self.getAllAtoms()
        masterlist = '{0}/masterlist/masterlist.ions'.format(self.Chianti_path)
        try:
            with open(masterlist, 'r') as f:
                lines = f.readlines()
        except:
            pn.config.error('Masterlist file not read', calling='addAllChianti')
            return None
        chianti_ions = [l[0:7].strip() for l in lines]
        chianti_ions = [i.capitalize().replace('_','') for i in chianti_ions if i[-1] != 'd']
        for cion in chianti_ions:
            if (cion not in atoms or replace_all) and cion not in ('H1', 'He1', 'He2'):
                try:
                    atfiles = self.getAllAvailableFiles(cion)
                    for atfile in atfiles:
                        if atfile.split('.')[-1] == 'chianti':
                            self.setDataFile(atfile)
                except:
                    pass
    
    def read_gsconf(self):
        try:
            gsconf = np.genfromtxt(ROOT_DIR + '/' + 'atomic_data/levels/gsconfs.dat', 
                                   names=['atom', 'gsconf'], dtype='U5, U5')
            self.gsconf = {gs['atom']:gs['gsconf'] for gs in gsconf}
        except:
            self.gsconf = {}
        
    def printPoem(self, yr=0):
        """
        Print one of Wiese et al. spectroscopic poems
        
        Parameters:
            - yr     publication year. If no valid year is given, random print
 
        """
        if (yr == 96):
            print("""********************************************
            
We have used a near limitless data source,
It is the Opacity Project, of course.
Its results are certainly fine
For most any LS-coupled line.


(Wiese, Fuhr & Deters 1996)""")            
        elif (yr == 66):
            print("""********************************************

If there is no other data source            
Use the Coulomb approximation, of course.
The results should be certainly fine
For any moderately or highly excited line.

(Wiese, Smith & Glennon 1966)""")
        else:
            rnd = np.random.rand()
            if (rnd < 0.5):
                self.printPoem(yr=66)
            else:
                self.printPoem(yr=96)


def extract_flt(str_):
    """
    Input is a bytes (py3) or a string (py2)
    Return the first float in str_, removing the no digit (and no dot) leading part of it.
    Ex:
    extract_flt('(123.00?') -> 123.00
    """
    res = ''
    if len(str_) > 0:
        if str_.decode()[0] in ('(', '['):
            str_ = str_[1:]
    for l in str_.decode():
        if l.isdigit() or l == '.':
            res += l
        else:
            break
    if res == '':
        return np.nan
    else:
        return float(res)

def readNIST(NISTfile,NLevels=None):
    """
    The NIST levels file must be obtained from this page:
        http://physics.nist.gov/PhysRefData/ASD/levels_form.html
    with the following options:
        Level Units: cm-1
        Format output: ASCII
        Display output: in its entirely
        Energy ordered
        Level Information: Principal configuration, Principal term, Level, J
        Bibliographic references
    """
    data = np.genfromtxt(NISTfile, comments = '-', 
                    delimiter = '|', names = 'conf, term, J, energy, ref', 
                    dtype=('U23', 'U9', 'U4', 'float', 'U10'), autostrip=True,
                    converters = {'energy':extract_flt})
    mask = data['J'] != ''
    data = data[mask]
    previous_ref = ''
    previous_conf = ''
    previous_term = ''
    for d in data:
        if d['conf'] == '':
            d['conf'] = previous_conf
        else:
            previous_conf = d['conf']
        if d['term'] == '':
            d['term'] = previous_term
        else:
            previous_term = d['term']
        if d['ref'] == '':
            d['ref'] = previous_ref
        else:
            previous_ref = d['ref']
            
        if '?' in d['J']:
            d['J'] = d['J'].split('?')[0]
        if d['J'] == '':
            d['J'] = 0.
        elif '/' in d['J']:
            d['J'] = '{0:.1f}'.format(float(d['J'].split('/')[0]) / float(d['J'].split('/')[1]))
        elif ',' in d['J']:
            d['J'] = '{0:.1f}'.format(float(d['J'].split(',')[0]))
        else:
            d['J'] = '{0:.1f}'.format(float(d['J']))
    data = data[data['energy'].argsort()]
    data = data.astype([('conf', 'U23'), ('term', 'U9'), ('J', 'float'), ('energy', 'float'), ('ref', 'U10')])
    if NLevels is not None:
        data = data[0:NLevels]
    return data


def getLevelsNIST(atom, NLevels = None):
    """
    Return a numpy.array containing the NIST data related to the levels for the given atom
    The keys of the recarry are: conf, term, J, energy, ref
    Example:
        data = getLevelsNIST('O3')
        energies = data['energy']
        
    """
    elem, spec = parseAtom(atom)
    elem = elem.lower()
    spec = int_to_roman(int(spec)).lower()
    atom_str = '{0}_{1}'.format(elem, spec)
    level_file = '{0}_levels.dat'.format(atom_str)
    file_ = ROOT_DIR + '/' + 'atomic_data/levels/{0}'.format(level_file)
    if os.path.isfile(file_):
        data = readNIST(file_, NLevels=NLevels)
        pn.log_.message('Reading energies and stat weights from {0}'.format(level_file), calling='getLevelsNIST')
        return data
    else:
        return None
    
def getNIST(elem, ion, fileout=None):
    """
    This will download the enegy levels from NIST and write them to a file
    example: getNIST('O',3) write the o_iii_levels.dat file
    """
    
    try:
        from selenium import webdriver
        from selenium.webdriver.support.ui import Select
    except:
        pn.log_.warn('selenium not installed', calling='get_NIST')
        return None
    import time
    atom_NIST = elem + ' ' + int_to_roman(int(ion))
    browser = webdriver.Firefox()
    try:
        browser.get('http://physics.nist.gov/PhysRefData/ASD/levels_form.html')
        Select(browser.find_element_by_name('format')).select_by_value('1')
        browser.find_element_by_css_selector("input[name='multiplet_ordered'][value='1']").click()
        browser.find_element_by_name('spectrum').send_keys(atom_NIST)
        browser.find_element_by_name('lande_out').click()
        browser.find_element_by_name('perc_out').click()
        browser.find_element_by_name('submit').click()
        time.sleep(1)
        tab = browser.find_element_by_tag_name('pre').text
    except:
        tab = None
    browser.close()
    
    if tab is not None:
        if fileout is None:
            fileout = '{}_{}_levels.dat'.format(elem.lower(), int_to_roman(int(ion)).lower())
        with open(fileout, 'w') as f:
            res = tab.split('\n')
            print_it = True
            i_line = 0
            for line in res[4:]:
                if '---' in line or '[' in line:
                    print_it = False
                if print_it:
                    f.write(line+'\n')
                    i_line += 1
        return i_line
    else:
        return None
            
    
def _atom_fits2ascii(filename):
    """
    Transform a *atom*.fits file into an *atom*.dat file.
    
    """
    strs = filename.split('/')[-1].split('_')
    elem = strs[0].capitalize()
    spec = roman_to_int(strs[1])
#    type = strs[2]
    pn.atomicData.setDataFile(filename.split('/')[-1])
    atom = pn.Atom(elem, spec)
    atom.printSources()
    print('')
    fileout = raw_input('Enter the name of the output file (between {0[0]}_{0[1]}_{0[2]}_ and .dat:'.format(strs))
    
    fout = open('{0[0]}_{0[1]}_{0[2]}_{1}.dat'.format(strs,fileout), 'w')
    fout.write('Aij\n')
#    str_ = ''
#    for i in np.arange(atom.NLevels):
#        str_ += '1/s '
#    str_ += '\n'
    str_ = '1/s ' * atom.NLevels + '\n'
    fout.write(str_)
    for a1 in atom.getA():
        str_ = ''
        for a2 in a1:
            str_ += '{0:10.7e}'.format(a2)
            str_ += ' '
        str_ += '\n'
        fout.write(str_)
    for item in atom.AtomData.AtomHeader.iteritems():
        if 'SPECTRUM' in item[0] or 'ATOM' in item[0] or 'GSCONFIG' in item[0]:
            fout.write('*** ' + item[0] + ' ' + str(item[1]) + '\n')
        if 'NOTE' in item[0]:
            if 'Energy' not in item[1]:
                N_Note = item[0].split('NOTE')[1]
                str_N_Note = str(N_Note.split('_')[0])
                fout.write('*** SOURCE' + str_N_Note + ' ' + atom.AtomData.AtomHeader.get('SOURCE'+str_N_Note)  + '\n')
                fout.write('*** ' + item[0] + ' ' + str(item[1]) + '\n')
    
    fout.close()
    
def _coll_fits2ascii(filename, overwrite=None):
    """
    Transform a *coll*.fits file into an *coll*.dat file.
    
    """
    strs = filename.split('/')[-1].split('_')
    elem = strs[0].capitalize()
    spec = roman_to_int(strs[1])
    pn.atomicData.setDataFile(filename.split('/')[-1])
    atom = pn.Atom(elem, spec)
    #atom.printSources()
    fileout = '{}.dat'.format(filename.split('.')[0])
    
    if overwrite is not True:
        if os.path.exists(fileout):
            erase = raw_input('{} exists, overwrite?'.format(fileout))
            if erase != 'y':
                return
    fout = open(fileout, 'w')
    fout.write('*** {0}{1} collision strengths data\n'.format(elem, strs[1].upper()))
    str_ = '0 0'
    for t in atom.getTemArray():
        str_ += ' {0:9.3e}'.format(t)
    str_ += '\n'
    fout.write(str_)
    for omega in atom.getOmegaArray():
        j = strExtract(omega, 'Omega(', '->')
        i = strExtract(omega, '->', ')')
        omegs = strExtract(omega, '[', ']')
        str_ = '{0} {1}'.format(i,j)
        for omeg in omegs.split():
            str_ += ' {0:9.3e}'.format(float(omeg))
        str_ += '\n'
        fout.write(str_)
        
    for item in atom.CollData.CollHeader.iteritems():
        if 'SPECTRUM' in item[0] or 'ATOM' in item[0] or 'GSCONFIG' in item[0]:
            fout.write('*** ' + item[0] + ' ' + str(item[1]) + '\n')
        if 'TUNIT1' == item[0]:
            fout.write('*** T_UNIT ' + str(item[1]) + '\n') 
        if 'SOURCE' in item[0]:
            N_Source = item[0].split('SOURCE')[1]
            str_N_Source = str(N_Source.split('_')[0])
            fout.write('*** ' + item[0] + ' ' + str(item[1]) + '\n')
            try:
                fout.write('*** NOTE' + str_N_Source + ' ' + atom.CollData.CollHeader.get('NOTE'+str_N_Source)  + '\n')
            except:
                fout.write('*** NOTE' + str_N_Source + ' collision strengths \n')
    print('{} done'.format(fileout))
    fout.close()
    
def print_stout_coll(Atom, file_):
    
    if type(Atom) is not pn.Atom:
        pn.log_.error('Atom must be a PyNeb.Atom object', calling='print_stout')
    
    temps = Atom.getTemArray(keep_unit=False)
    n_temp = len(temps)
    omegas = Atom.getOmegaArray()
    n_levels = omegas.shape[0]
    with open(file_, 'w') as f:
        f.write('11 10 14\n')
        f.write('TEMP {} \n'.format(' '.join(['{:.0f}'.format(t) for t in temps])))
        for i in range(n_levels):
            for j in range(n_levels):
                if omegas[i,j].sum() > 0.0:
                    f.write('CS ELECTRON {} {} {} \n'.format(i+1, j+1, ' '.join([str(o) for o in omegas[i,j]])))
        f.write('***************\n')
        f.write('Refs:\n')
        for s in Atom.CollData.getSources():
            f.write('{} \n'.format(s))





import os, sys
import pyneb as pn
import numpy as np
from pyneb.utils.misc import int_to_roman, roman_to_int
try:
    import pyfits
except:
    try:
        import astropy.io.fits as pyfits
    except:
        pn.log_.error('pyfits not installed')
from physics import sym2name, gsFromAtom

    
def _updateHeader(hdu, input_file, content, elem, spec):
    """
    Update the header of the output fits file

    Parameters: 
        - hdu          pyfits file
        - input_file   name of the ascii files containing the original data
        - content      type of physical quantities contained in the file
        - elem         symbol of the selected element
        - spec         ionization stage in spectroscopic notation (I = 1, II = 2, etc.)
        
    """
    hdu.header.update('ORIGIN', 'File {0}, PyNeb v{1}'.format(input_file, pn.__version__))
    hdu.header.update('CONTENT', content)
    hdu.header.update('ATOM', sym2name[elem.capitalize()])
    hdu.header.update('SPECTRUM', roman_to_int(spec))
    atom = elem.capitalize() + str(roman_to_int(spec))
    gs = gsFromAtom(atom)
    hdu.header.update('GSCONFIG', gs)
    

def writeAtom(input_file):
    """ 
    Write the Es and As read from an ascii file into a fits file in PyNeb format.
    The bibliographic references are read from comment lines in the file.
        
    Parameters:
        - input_file    name of the ascii input file containing the energy levels (in 1/A or eV) and the Aij,
                        in the format elem_spec_atom_REF-STRING.dat

    Example: 
    writeAtom('o_iii_atom_AB01-CD02-EFG03.dat')

    """
    if not os.path.exists(input_file):
        pn.log_.error('{0} not found.'.format(input_file), calling='writeAtom')
        sys.exit(0)
    data_type = input_file.split(".")[0].split("_")[2]
    if (data_type != 'atom'):
        pn.log_.error('The name of {0} does not conform to the prescribed format.'.format(input_file), calling='writeAtom')
        sys.exit(0)
    elem = input_file.split(".")[0].split("_")[0]
    spec = input_file.split(".")[0].split("_")[1]
    ref = input_file.split(".")[0].split("_")[-1]
    out_fits = '{0}_{1}_{2}_{3}.fits'.format(elem.lower(), spec.lower(), data_type, ref)
    atom_dat = []

    # Read data from ascii file
    #
    # Read energies and stat weights 
    f = open(input_file)
    units = f.readlines()[1].split()[0]
    all_data = np.genfromtxt(input_file, comments='***', skip_header=2, unpack=True)
    # Read Es
    energy = all_data[0, :]
    NLevels = len(energy)
    if units[0] is 'eV': 
        energy = energy / (pn.CST.RYD_EV * pn.CST.RYD_ANG)
    # Read statistical weights
    stat_weight = all_data[1, :]
    atom_dat.append(pyfits.Column(name='Energy', format='1D', unit='1/Ang', array=energy))
    atom_dat.append(pyfits.Column(name='Stat_Weight', format='1J', unit='none', array=stat_weight))    
    # Read As
    for j in np.arange(NLevels - 1):
        lev_j = j + 1
        fieldName = 'A(*->{0})'.format(lev_j)
        atom_dat.append(pyfits.Column(name=fieldName, format='1D', unit='1/sec', array=all_data[1 + lev_j, :]))

    # Create fits file and start adding info to the header
    atom_hdu = pyfits.new_table(atom_dat)   
            
    # Primary header (extension 0)
    hdu = pyfits.PrimaryHDU()
    hdu.header.update('FILENAME', out_fits)
    hdu.header.update('NEXTEND', 1)
 
    # Write references in the header
    refs = []
    for line in open(input_file):
        if '***' in line:
            text = line.strip("***").split("'")
            if ('SOURCE' in text[0]) or ('NOTE' in text[0]):
                refs.append((text[0], text[1]))

    _updateHeader(atom_hdu, input_file, 'Energy levels and transition probabilities', elem, spec)
    _updateHeader(hdu, input_file, 'Energy levels and transition probabilities', elem, spec)
    atom_hdu.header.update('N_LEVELS', NLevels)
    for item in refs:
        atom_hdu.header.update(item[0], item[1])

    hdulist = pyfits.HDUList([hdu])
    hdulist.append(atom_hdu)
    hdulist.writeto(out_fits, clobber=True)


def writeColl(input_file, temp_in_cols = True, chebOrder = None):
    """ 
    Write the Omegas read from an ascii file into a fits file in PyNeb format.
    The bibliographic references are read from comment lines in the file.
        
    Parameters:
        - input_file    name of the ascii input file containing the Omegas,
                        in the format elem_spec_coll_REF-STRING.dat
        - temp_in_cols  If True, each column of the input file corresponds to a given Te, 
                         otherwise each rows corresponds to a given Te (a transposition is 
                         thus applied to the data)
    Example: 
    writeColl('o_iii_coll_AB01-CD02-EFG03.dat')

    """
    def _omegaArray(lev_j, lev_i):
        """
        Return the array of collision strength values for given levels
 
        """       
        ij = (cs[:, 0] == lev_j) & (cs[:, 1] == lev_i)
        if ij.sum() == 0.:
            return np.zeros(n_temp)
        else:
            return cs[ij, 2::].reshape(n_temp)

    if not os.path.exists(input_file):
        pn.log_.error('{0} not found.'.format(input_file), calling='writeColl')
        sys.exit(0)
    data_type = input_file.split(".")[0].split("_")[2]
    if (data_type != 'coll'):
        pn.log_.error('The name of {0} does not conform to the prescribed format.'.format(input_file), calling='writeColl')
        sys.exit(0)
    elem = input_file.split(".")[0].split("_")[0]
    spec = input_file.split(".")[0].split("_")[1]
    ref = input_file.split(".")[0].split("_")[-1]
    out_fits = '{0}_{1}_{2}_{3}.fits'.format(elem.lower(), spec.lower(), data_type, ref)
    coll_dat = []

    # Read Omegas
    cs = np.genfromtxt(input_file, comments='***')
    if not temp_in_cols:
        cs = cs.transpose()
    n_temp = np.size(cs[0]) - 2
    temArray = cs[0, 2::]
    NLevels = int(np.max(cs[:, 1]))
    if chebOrder is None:
        chebOrder = n_temp
    coll_dat.append(pyfits.Column(name='Te', format='1E', unit='log(K)', array=temArray)) # Change if linear T used (see below)
    for j in np.arange(NLevels - 1):
        lev_j = j + 1
        lev_i = lev_j + 1
        while (lev_i < NLevels + 1):
            fieldName = 'Omega({0}->{1})'.format(int(lev_i), int(lev_j))
            omegaArray = _omegaArray(lev_j, lev_i)
            coll_dat.append(pyfits.Column(name=fieldName, format='1E', unit=chebOrder, array=omegaArray))
            lev_i += 1
    coll_hdu = pyfits.new_table(coll_dat)        
            
    # Primary header (extension 0)
    hdu = pyfits.PrimaryHDU()
    hdu.header.update('FILENAME', out_fits)
    hdu.header.update('NEXTEND', 1)
 
    # Write references in the header
    refs = []
    for line in open(input_file):
        if '***' in line:
            text = line.strip("***").split("'")
            if ('SOURCE' in text[0]) or ('NOTE' in text[0]):
                refs.append((text[0], text[1]))

    _updateHeader(coll_hdu, input_file, 'Collision strengths', elem, spec)
    _updateHeader(hdu, input_file, 'Collision strengths', elem, spec)
    coll_hdu.header.update('N_LEVELS', NLevels)
    for item in refs:
        coll_hdu.header.update(item[0], item[1])

    hdulist = pyfits.HDUList([hdu])
    hdulist.append(coll_hdu)
    hdulist.writeto(out_fits, clobber=True)


class Hdr(object):
    """
    Explore and update headers of fits files
    
    Parameters:
        - fitsFile  name of the fits file
        - ext       file extension
    
    Examples:
        f=hdr('o_iii.fits', 2)
        f.show()
        f.up('ATOM', 'kriptonyte')
        f.up('NATURE', 'invented')
        f.save
        
    """    
    def __init__(self, fitsFile, ext):
        self.fitsFile = fitsFile
        self.hdu = pyfits.open(fitsFile, ignore_missing_end=True)
        self.ext = ext
    

    def show(self):
        print self.hdu[self.ext].header
    

    def up(self, key, value):
        """ This will be changed to header.set at some point"""
        self.hdu[self.ext].header.update(key, value)
        

    def delete(self, key):
        del self.hdu[self.ext].header[key]


    def cheb(self, orderFile):    
        """ 
        Update the Chebyshev order of a coll fits file.
        
        Parameters:
            - orderFile    ascii file with upper level, lower level, order
        
        """    
        j_order, i_order, order = np.genfromtxt(orderFile, comments='***', usecols=(0, 1, 2), unpack=True)    
        NLevels = np.max(j_order)
        order_array = np.zeros([NLevels, NLevels])
        for index in range(len(j_order)):
            order_array[int(j_order[index] - 1), int(i_order[index] - 1)] = int(order[index])
            
        # Iterate over the keys to associate each unit with its Omega(*->*) field
        for key in self.hdu[self.ext].header.keys():
            if 'TTYPE' in key:
                num = key.strip('TTYPE')
                unit = 'TUNIT' + num
                field_name = self.hdu[self.ext].header[key]
                # The try is to skip the Te field
                try:
                    lev_j = field_name.split('->')[0].split('Omega(')[1]
                    lev_i = field_name.split('->')[1].split(')')[0]
                    j = int(lev_j) - 1
                    i = int(lev_i) - 1
                    self.up(unit, order_array[j, i])
                except:
                    pass


    def save(self):
        self.hdu.writeto(self.fitsFile, clobber=True)

# Test
# writeAtom('o_ii_atom_TEST.dat')
# writeColl('o_ii_coll_TEST.dat')



import os
from .logging import my_logging

class _Config(object):
    """
    This is the place where to put stuff that any module may need, a kind of COMMON.
    An instantiation is done in the main __init__ file using the "config" name.

    """


    def __init__(self):
        """
        Define variables that will be known from everywhere:
        - INSTALLED: a dictionary whose keys are the libraries that PyNeb may need
            e.g. 'plt' for matplotlib.pyplot, 'scipy' for scipy.interpolate. and whose values
            are Boolean
        
        - Datafile: a dictionary holding the HI atom, which can be used intensively in Atom.getIonAbundance
        
        - pypic_path: This is only set when needed (basically by getEmisGridDict).
            The default value for the pypic_path parameter of getEmisGridDict().
            The first try is ./.pypics: if it cannot be created, or it exists but it is not writable, 
            /tmp/pypics is tried; if it cannot be created, or it exists but it is not writable, the path
            is set to None and a pypic_path value shall be provided to getEmisGridDict().
        
        Parameters: none
        
        """
        self.log_ = my_logging()
        self.calling = '_Config'
        
        self.INSTALLED = {}
        try:
            import matplotlib.pyplot as plt
            self.INSTALLED['plt'] = True
        except:
            self.INSTALLED['plt'] = False
            self.log_.message('matplotlib not available', calling=self.calling)
        try:     
            from scipy import interpolate
            self.INSTALLED['scipy'] = True
        except:
            self.INSTALLED['scipy'] = False
            self.log_.message('scipy not available or interpolate not in scipy', calling=self.calling)
        try:
            import multiprocessing as mp
            self.INSTALLED['mp'] = True
            self.Nprocs = mp.cpu_count()
        except:
            self.INSTALLED['mp'] = False
            self.log_.message('multiprocessing not available', calling=self.calling)
            self.Nprocs = 1
        if 'XUVTOP' in os.environ:
            self.INSTALLED['Chianti'] = True
            try:
                xuvtop = os.environ['XUVTOP']
                vFileName = os.path.join(xuvtop, 'VERSION')
                vFile = open(vFileName)
                versionStr = vFile.readline()
                vFile.close()
                self.Chianti_version = versionStr.strip()
                self.Chianti_version_main = self.Chianti_version.split('.')[0]
            except:
                self.Chianti_version = None
                self.Chianti_version_main = None
        else:
            self.INSTALLED['Chianti'] = False
            self.Chianti_version = None
            self.Chianti_version_main = None
        if 'STOUT_DIR' in os.environ:
            self.INSTALLED['Stout'] = True
            self.Stout_dir = os.environ['STOUT_DIR']
        else:
            self.INSTALLED['Stout'] = False
            self.Stout_dir = None
        try:
            import astropy.io.fits as pyfits
            self.INSTALLED['pyfits from astropy'] = True
            self.INSTALLED['pyfits'] = True
        except:
            self.INSTALLED['pyfits from astropy'] = False 
            try:
                import pyfits
                self.INSTALLED['pyfits'] = True
            except:
                self.INSTALLED['pyfits'] = False
        try:
            import h5py
            self.INSTALLED['h5py'] = True
        except:
            self.INSTALLED['h5py'] = False
        try:
            from astropy.table import Table
            self.INSTALLED['astropy Table'] = True
        except:
            self.INSTALLED['astropy Table'] = False
        try:
            import cvxopt
            self.INSTALLED['cvxopt'] = True
        except:
            self.INSTALLED['cvxopt'] = False
        
        self.DataFiles = {}
            
        self.unuse_multiprocs()
        
        self.kappa = None
        self.set_noExtrapol(False)
                    
        self.__pypic_path = None
        
        self.vactoair_low_wl = 2000. # UV in vacuum
        self.vactoair_high_wl = 1e30 # no upper limit, IR in air!!!
                
    def set_noExtrapol(self, value):
        self._noExtrapol = bool(value)
        
    def get_noExtrapol(self):
        return self._noExtrapol
        
    def _get_PypicPath(self):
        pypic_path = self.__pypic_path
        if pypic_path is None:
            pypic_path = './pypics/'
            if os.path.exists(pypic_path):
                if not os.access(pypic_path, os.W_OK):
                    self.log_.warn('Directory {0} not writable'.format(pypic_path),
                                      calling=self.calling)
                    pypic_path = None
            else:
                try:
                    os.makedirs(pypic_path)
                    self.log_.message('Directory {0} created'.format(pypic_path),
                                      calling=self.calling)
                except:
                    pypic_path = None
            if pypic_path is None:
                pypic_path = '/tmp/pypics/'
                if os.path.exists(pypic_path):
                    if not os.access(pypic_path, os.W_OK):
                        self.log_.warn('Directory {0} not writable'.format(pypic_path),
                                          calling=self.calling)                                   
                        pypic_path = None 
                else:
                    try:
                        os.makedirs(pypic_path)
                        self.log_.message('Directory {0} created'.format(pypic_path),
                                          calling=self.calling)
                    except:
                        pypic_path = None
        else:
            if os.path.exists(pypic_path):
                if not os.access(pypic_path, os.W_OK):
                    self.log_.warn('Directory {0} not writable'.format(pypic_path),
                                      calling=self.calling)
                    pypic_path = None
            else:
                try:
                    os.makedirs(pypic_path)
                    self.log_.message('Directory {0} created'.format(pypic_path),
                                      calling=self.calling)
                except:
                    self.log_.warn('Unable to create directory {0}'.format(pypic_path),
                                      calling=self.calling)
                    
                    pypic_path = None
            
        self.log_.message('pypic_path set to {0}'.format(pypic_path),
                                          calling=self.calling)
        self.__pypic_path = pypic_path
        
        return self.__pypic_path
    
    def _set_PypicPath(self, value):            
        self.__pypic_path = value
        test = self.pypic_path
        
    pypic_path = property(_get_PypicPath, _set_PypicPath, None, None)
            
    def use_multiprocs(self):
        self._use_mp = True
    
    def unuse_multiprocs(self):
        self._use_mp = False
        
        
            


import sys
import re
import pdb
import os

IS_PYTHON3 = sys.version_info[0] >= 3

def _parser(tokens, version=None):
    # Parse the full edit descriptors
    eds = _parse_tokens(tokens, reversion=False, version=None)
    # Parse the edit descriptors used for the reversion of format control 
    # (F95 format 12.2.2)
    reversion_eds = _parse_tokens(tokens, reversion=True, version=None)
    return eds, reversion_eds

def _parse_tokens(tokens, reversion=False, version=None):
    # Remove outer parens is there are any
    tokens = _remove_outer_parens(tokens)
    # Get only the reversion tokens
    if reversion == True:
        tokens = _get_reversion_tokens(tokens)
    # First expand the parentheses
    tokens = _expand_parens(tokens)
    # Split on commas
    token_sets = _split_on_commas(tokens)
    # Split on ED9 (i.e. :)
    token_sets = _split_on_ed9(token_sets)
    # Split the ED10 (i.e. /)
    token_sets = _split_on_ed10(token_sets)
    # Split the ED8 (i.e. P edit descriptors)
    token_sets = _split_on_ed8(token_sets)
    # Process each set of edit descriptors
    eds = []
    for token_set in token_sets:
        # Assume first edit descriptor is the one to process
        ed_type = None
        ed_value = None
        for token in token_set:
            if token.type in ['ED1', 'ED2', 'ED3', 'ED4', 'ED5', 'ED6', 'ED7', 'ED8', 'ED9', 'ED10', 'QUOTED_STRING']:
                ed_type = token.type
                ed_value = token.value
                break
        # TODO: something more responsible here ...
        if ed_type is None:
            continue
        # If repeatable and first token is repeat number then cache
        repeat = None
        if ed_value in REPEATABLE_EDS and (token_set[0].type in ['NZUINT', 'UINT']):
            repeat = token_set[0].value
            token_set = token_set[1:]
        # Process the edit descriptor
        if ed_type == 'QUOTED_STRING':
            ed = _read_quoted_string(token_set)
        elif ed_type == 'ED1':
            ed = _read_ed1(token_set)
        elif ed_type == 'ED2':
            ed = _read_ed2(token_set)
        elif ed_type == 'ED3':
            ed = _read_ed3(token_set)
        elif ed_type == 'ED4':
            ed = _read_ed4(token_set)
        elif ed_type == 'ED5':
            ed = _read_ed5(token_set)
        elif ed_type == 'ED6':
            ed = _read_ed6(token_set)
        elif ed_type == 'ED7':
            ed = _read_ed7(token_set)
        elif ed_type == 'ED8':
            ed = _read_ed8(token_set)
        elif ed_type == 'ED9':
            ed = _read_ed9(token_set)
        elif ed_type == 'ED10':
            ed = _read_ed10(token_set)
        else:
            raise InvalidFormat('Could not identify edit descriptor in sequence $s' % str(token_set))
        # If there is a repeat number cached, then apply
        if repeat is not None:
            ed.repeat = repeat
        # Add to the list
        eds.append(ed)
    return eds


# Functions that group the tokens into sets

def _expand_parens(tokens):
    new_tokens = []
    get_tokens = iter(tokens)
    for t0 in get_tokens:
        if t0.type != 'LEFT_PARENS':
            new_tokens.append(t0)
        else:
            # Read in all tokens in subformat and recurse back to self
            paren_tokens = []
            nesting = 1
            while nesting > 0:
                try:
                    if IS_PYTHON3:
                        t1 = next(get_tokens)
                    else:
                        t1 = get_tokens.next()
                except StopIteration:
                    raise InvalidFormat('Open parens in format')
                if t1.type == 'LEFT_PARENS':
                    nesting = nesting + 1
                elif t1.type == 'RIGHT_PARENS':
                    nesting = nesting - 1
                paren_tokens.append(t1)
            # Remove last right paren
            paren_tokens = paren_tokens[:-1]
            # If repeated, then repeat the tokens accordingly
            if (len(new_tokens) > 0) and (new_tokens[-1].type in ['NZUINT', 'UINT']):
                repeat = new_tokens[-1].value
                # Remove the repeat NZUINT, UINT
                new_tokens = new_tokens[:-1]
                new_tokens.extend(repeat * (_expand_parens(paren_tokens) + [Token('COMMA', None)]))
            else:
                new_tokens.extend(_expand_parens(paren_tokens))
    return new_tokens


def _split_on_commas(tokens):
    token_sets = []
    set_buff = []
    for t0 in tokens:
        if t0.type == 'COMMA':
            token_sets.append(set_buff)
            set_buff = []
        else:
            set_buff.append(t0)
    token_sets.append(set_buff)
    return token_sets


def _split_on_ed9(token_sets):
    '''Splits on :'''
    new_token_sets = []
    for token_set in token_sets:
        if 'ED9' not in [t.type for t in token_set]:
            new_token_sets.append(token_set)
        else:
            buff = []
            for token in token_set:
                if token.type == 'ED9':
                    if len(buff) > 0:
                        new_token_sets.append(buff)
                        buff = []
                    new_token_sets.append([token])
                else:
                    buff.append(token)
            if len(buff) > 0:
                new_token_sets.append([token])
    return new_token_sets


def _split_on_ed10(token_sets):
    '''Splits on /'''
    new_token_sets = []
    for token_set in token_sets:
        # May have a repeat on the slash if preceded by a comma
        if (len(token_set) > 2) and ((token_set[0].type in ['UINT', 'NZUINT']) and (token_set[1].type == 'ED10')):
            new_token_sets.append(token_set[:2])
            token_set = token_set[2:]
        buff = []
        for token in token_set:
            if token.type == 'ED10':
                if len(buff) > 0:
                    new_token_sets.append(buff)
                    buff = []
                new_token_sets.append([token])
            else:
                buff.append(token)
        if len(buff) > 0:
            new_token_sets.append(buff)
    return new_token_sets


def _split_on_ed8(token_sets):
    '''Splits on ED8 (i.e. P edit descriptors)'''
    new_token_sets = []
    for token_set in token_sets:
        # Append to new list if no ED8
        if 'ED8' not in [t.type for t in token_set]:
            new_token_sets.append(token_set)
        # Otherwise split string on ED9
        elif (token_set[0].type in ['INT', 'UINT', 'NZUINT']) and (token_set[1].type == 'ED8'):
            new_token_sets.append(token_set[:2])
            new_token_sets.append(token_set[2:])
        else:
            raise InvalidFormat('P edit descriptor in invalid position')
    return new_token_sets

# Function to extract only the tokens for the reversion of control

def _get_reversion_tokens(tokens):
    reversion_tokens = []
    # Easier to work backwards
    nesting = None
    for token in tokens[::-1]:
        # End of loop condition
        if (nesting is not None) and (nesting < 1):
            # Parens may have a repeat number
            if token.type in ['UINT', 'NZUINT']:
                reversion_tokens.append(token)
            break
        # Read till the first right parens
        if token.type == 'RIGHT_PARENS':
            if nesting is None:
                nesting = 1
            else:
                nesting = nesting + 1
        elif token.type == 'LEFT_PARENS':
            if nesting is None:
                raise InvalidFormat('Unbalanced parens in format')
            else:
                nesting = nesting - 1
        reversion_tokens.append(token)
    # Tokens are in reverse order
    reversion_tokens.reverse()
    return reversion_tokens

# The functions that read particular edit descriptors sequences

def _read_quoted_string(tokens):
    # Of form X only
    type_string = ",".join([t.type for t in tokens])
    if type_string != "QUOTED_STRING":
        raise InvalidFormat('Token %s has invalid neighbouring token' % tokens[0])
    ed = QuotedString()  
    ed.char_string = tokens[0].value
    return ed

def _read_ed1(tokens):
    # Of form X only
    type_string = ",".join([t.type for t in tokens])
    if type_string != "ED1":
        raise InvalidFormat('Token %s has invalid neighbouring token' % tokens[0])
    ed = get_edit_descriptor_obj(tokens[0].value)   
    return ed

def _read_ed2(tokens):
    # Of form nX only
    type_string = ",".join([t.type for t in tokens])
    if type_string != "NZUINT,ED2":
        raise InvalidFormat('Token %s has invalid neighbouring token' % tokens[0])
    ed = get_edit_descriptor_obj(tokens[1].value)
    ed.num_chars = tokens[0].value
    return ed

def _read_ed3(tokens):
    # Of form Xn only
    type_string = ",".join([t.type for t in tokens])
    if type_string != "ED3,NZUINT":
        raise InvalidFormat('Token %s has invalid neighbouring token' % tokens[0])
    ed = get_edit_descriptor_obj(tokens[0].value)
    # L edit descriptor has a width rather than num_chars
    if hasattr(ed, 'width'):
        ed.width = tokens[1].value
    else:
        ed.num_chars = tokens[1].value
    return ed

def _read_ed4(tokens):
    # Of form X or Xn
    type_string = ",".join([t.type for t in tokens])
    if type_string in ["ED4", "ED4,NZUINT"] or \
        (ALLOW_ZERO_WIDTH_EDS and (type_string == "ED4,UINT")):
        ed = get_edit_descriptor_obj(tokens[0].value)
        if len(tokens) > 1:
            ed.width = tokens[1].value
    else:
        raise InvalidFormat('Token %s has invalid neighbouring token' % tokens[0])
    return ed

def _read_ed5(tokens):
    # Of form Xn.m only
    type_string = ",".join([t.type for t in tokens])
    if type_string in ["ED5,NZUINT,DOT,UINT", "ED5,NZUINT,DOT,NZUINT"] or \
      (ALLOW_ZERO_WIDTH_EDS and (type_string in \
      ["ED5,UINT,DOT,UINT", "ED5,UINT,DOT,NZUINT"])):
        ed = get_edit_descriptor_obj(tokens[0].value)
        ed.width = tokens[1].value
        ed.decimal_places = tokens[3].value
    else:
        raise InvalidFormat('%s has invalid neighbouring token' % tokens[0])
    return ed

def _read_ed6(tokens):
    # Of form Xn or Xn.m
    type_string = ",".join([t.type for t in tokens])
    if type_string == "ED6,NZUINT" or \
        (ALLOW_ZERO_WIDTH_EDS and (type_string == "ED6,UINT")):
        ed = get_edit_descriptor_obj(tokens[0].value)
        ed.width = tokens[1].value
        ed.min_digits = None
    elif type_string in ["ED6,NZUINT,DOT,UINT", "ED6,NZUINT,DOT,NZUINT"] or \
      (ALLOW_ZERO_WIDTH_EDS and (type_string in \
      ["ED6,UINT,DOT,UINT", "ED6,UINT,DOT,NZUINT"])):
        ed = get_edit_descriptor_obj(tokens[0].value)
        ed.width = tokens[1].value
        ed.min_digits = tokens[3].value
    else:
        raise InvalidFormat('%s has invalid neighbouring token' % tokens[0])
    return ed

def _read_ed7(tokens):
    # Of form Xn.m or Xn.mEe
    type_string = ",".join([t.type for t in tokens])
    if type_string in ["ED7,NZUINT,DOT,UINT", "ED7,NZUINT,DOT,NZUINT"] or \
      (ALLOW_ZERO_WIDTH_EDS and (type_string in \
      ["ED7,UINT,DOT,UINT", "ED7,UINT,DOT,NZUINT"])):
        ed = get_edit_descriptor_obj(tokens[0].value)
        ed.width = tokens[1].value
        ed.decimal_places = tokens[3].value
        ed.exponent = None
    elif type_string in ['ED7,NZUINT,DOT,NZUINT,ED7,NZUINT', \
            'ED7,NZUINT,DOT,NZUINT,ED7,UINT', \
            'ED7,NZUINT,DOT,NZUINT,ED7,INT', \
            'ED7,NZUINT,DOT,UINT,ED7,NZUINT', \
            'ED7,NZUINT,DOT,UINT,ED7,UINT', \
            'ED7,NZUINT,DOT,UINT,ED7,INT'] or \
        (ALLOW_ZERO_WIDTH_EDS and (type_string in \
            ['ED7,UINT,DOT,NZUINT,ED7,NZUINT', \
            'ED7,UINT,DOT,NZUINT,ED7,UINT', \
            'ED7,UINT,DOT,NZUINT,ED7,INT', \
            'ED7,UINT,DOT,UINT,ED7,NZUINT', \
            'ED7,UINT,DOT,UINT,ED7,UINT', \
            'ED7,UINT,DOT,UINT,ED7,INT'])):
        ed = get_edit_descriptor_obj(tokens[0].value)
        ed.width = tokens[1].value
        ed.decimal_places = tokens[3].value
        ed.exponent = tokens[5].value
    else:
        raise InvalidFormat('%s has invalid neighbouring token' % tokens[0])
    return ed

def _read_ed8(tokens):
    # Of form kX only, where k is a signed integer, may omit comma if followed
    # by Type 5 or 7 edit descriptor
    type_string = ",".join([t.type for t in tokens])
    if type_string in ["NZUINT,ED8", "UINT,ED8", "INT,ED8"]:
        ed = get_edit_descriptor_obj(tokens[1].value)
        ed.scale = tokens[0].value
    else:
        raise InvalidFormat('%s has invalid neighbouring token' % tokens[0])
    return ed

def _read_ed9(tokens):
    # Of form X only, may omit comma either side
    type_string = ",".join([t.type for t in tokens])
    if type_string == "ED9":
        ed = get_edit_descriptor_obj(tokens[0].value)
    else:
        raise InvalidFormat('%s has invalid neighbouring token' % tokens[0])
    return ed

def _read_ed10(tokens):
    # Of form X only, may omit following comma and leading comma if no repeat
    type_string = ",".join([t.type for t in tokens])
    if type_string == "ED10":
        ed = get_edit_descriptor_obj(tokens[0].value)
    else:
        raise InvalidFormat('%s has invalid neighbouring token' % tokens[0])
    return ed


# Functions that pre-process the token list

def _remove_outer_parens(tokens):
    # Finally, remove outer parens is there are any
    if (tokens[0].type == 'LEFT_PARENS') and (tokens[-1].type == 'RIGHT_PARENS'):
        tokens = tokens[1:-1]
    return tokens


# Run some tests if run as a script

#if __name__ == '__main__':
#    import doctest
#    import os
#    from _lexer import lexer
#    globs = {'lexer' : lexer, 'parser' : parser}
#    # Need to normalize whitespace since pasting into VIM converts tabs to
#    # spaces
#    doctest.testfile(os.path.join('tests', 'parser_test.txt'), \
#        globs=globs, optionflags=doctest.NORMALIZE_WHITESPACE)

# Some lexer tokens to look out for
DIGITS = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '0']
SIGNS = ['+', '-']
COMMA = [',']
DOT = ['.']
WHITESPACE = [' ', '\t', '\n']
QUOTE_CHARS = ['"', "'"]
DOUBLE_EDIT_DESCRIPTORS = ['EN', 'ES', 'TR', 'TL', 'BN', 'BZ', 'SP', 'SS']
SINGLE_EDIT_DESCRIPTORS = ['A', 'B', 'D', 'E', 'F', 'G', 'I', 'L', 'O', 'P', 'S', 'T', 'X', 'Z', ':', '/']
H_EDIT_DESCRIPTOR = ['H']
LEFT_PARENS = ['(']
RIGHT_PARENS = [')']
COLON = ':'
SLASH = '/'


def _lexer(format):
    '''Lex the FORTRAN format statement into tokens'''
    tokens = []
    s = -1
    h_chars = None
    while True:
        # Get the next set of characters
        s = s + 1
        c0, c1, c2 = _get_chars(format, s)
        # If at end of format, end it all - aieee!
        if c0 is None:
            break
        # Read in an H edit descriptor string
        elif h_chars is not None:
            buff = format[s:s+h_chars]
            tokens.append(Token('QUOTED_STRING', buff))
            s = s + (h_chars - 1)
            h_chars = None
        # Skip whitespace
        elif c0 in WHITESPACE:
            continue
        # Read in a quoted string
        elif c0 in QUOTE_CHARS:
            buff = ''
            delim = c0
            while True:
                s = s + 1
                c0, c1, c2 = _get_chars(format, s)
                # Check if an escaped delimiter
                if (c0 == delim) and (c1 == delim):
                    s = s + 1
                    buff = buff + delim
                elif (c0 == delim):
                    break
                elif c0 is None:
                    # Premature end of format
                    raise InvalidFormat('Premature end of quoted string in format')
                else:
                    buff = buff + c0
            tokens.append(Token('QUOTED_STRING', buff))
        # Read in an integer
        elif c0 in DIGITS + SIGNS:
            # Check sign followed by at least one digit
            if (c0 in SIGNS) and (c1 not in DIGITS):
                # TODO: Is whitesapce allowed between sign and digit?
                raise InvalidFormat("Orphaned sign '%s' with no digits at position %d" % (c0, s))
            buff = c0
            while True:
                s = s + 1
                c0, c1, c2 = _get_chars(format, s)
                if (c0 not in DIGITS) or (c0 is None):
                    break
                else:
                    buff = buff + c0
            s = s - 1
            val = int(buff)
            if buff[0] in SIGNS:
                tokens.append(Token('INT', val))
            elif val == 0:
                tokens.append(Token('UINT', val))
            else:
                tokens.append(Token('NZUINT', val))
        # Read in a comma
        elif c0 in COMMA:
            tokens.append(Token('COMMA', None))
        # Read in a dot
        elif c0 in DOT:
            tokens.append(Token('DOT', None))
        # Read in double lettered edit descriptors
        elif (c1 is not None) and ((c0 + c1).upper() in DOUBLE_EDIT_DESCRIPTORS):
            ed_type = _get_ed_type((c0 + c1).upper())
            tokens.append(Token(ed_type, (c0 + c1).upper()))
            s = s + 1
        # Read in an H edit descriptor
        elif c0.upper() in H_EDIT_DESCRIPTOR:
            if (len(tokens) > 0) and (tokens[-1].type in ('NZUINT', 'UINT')):
                h_chars = tokens[-1].value
                tokens = tokens[:-1]
            else:
                raise InvalidFormat("Missing H descriptor number argument at position %d" % s)
        # Read in single lettered edit descriptors
        elif c0.upper() in SINGLE_EDIT_DESCRIPTORS:
            ed_type = _get_ed_type(c0.upper())
            tokens.append(Token(ed_type, c0.upper()))
        # Read in left parens
        elif c0 in LEFT_PARENS:
            tokens.append(Token('LEFT_PARENS', None))
        # Read in right parens
        elif c0 in RIGHT_PARENS:
            tokens.append(Token('RIGHT_PARENS', None))
        else:
            raise InvalidFormat('Character %s not recognised at position %d' % (c0, s))
    return tokens

def _get_ed_type(ed_string):
    if ed_string in ED1:
        ed_type = 'ED1'
    elif ed_string in ED2:
        ed_type = 'ED2'
    elif ed_string in ED3:
        ed_type = 'ED3'
    elif ed_string in ED4:
        ed_type = 'ED4'
    elif ed_string in ED5:
        ed_type = 'ED5'
    elif ed_string in ED6:
        ed_type = 'ED6'
    elif ed_string in ED7:
        ed_type = 'ED7'
    elif ed_string in ED8:
        ed_type = 'ED8'
    elif ed_string in ED9:
        ed_type = 'ED9'
    elif ed_string in ED10:
        ed_type = 'ED10'
    else:
        ed_type = None
    return ed_type

def _get_chars(format, s):
    try:
        c0 = format[s]
    except IndexError:
        c0 = None
    try:
        c1 = format[s+1]
    except IndexError:
        c1 = None
    try:
        c2 = format[s+2]
    except IndexError:
        c2 = None
    return (c0, c1, c2)


class InvalidFormat(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class Token(object):
    def __init__(self, type, value):
        self.type = type
        self.value = value
    def __repr__(self):
        return "\n  Token: type=%s,\tvalue=%s" % (self.type, str(self.value))

# Do some testing when run as a module

#if __name__ == '__main__':
#    import doctest
#    import os
#    globs = {'lexer' : lexer}
#    # Need to normalize whitespace since pasting into VIM converts tabs to
#    # spaces
#    doctest.testfile(os.path.join('tests', 'lexer_test.txt'), \
#        globs=globs, optionflags=doctest.NORMALIZE_WHITESPACE)

class InvalidFormat(Exception):
    pass

def get_edit_descriptor_obj(name):
    '''Returns a new object instance from a string'''
    name = name.upper()
    if name == 'A':
        return A()
    elif name == 'B':
        return B()
    elif name == 'BN':
        return BN()
    elif name == 'BZ':
        return BZ()
    elif name == ':':
        return Colon()
    elif name == 'D':
        return D()
    elif name == 'E':
        return E()
    elif name == 'EN':
        return EN()
    elif name == 'ES':
        return ES()
    elif name == 'F':
        return F()
    elif name == 'G':
        return G()
    elif name == 'H':
        return H()
    elif name == 'I':
        return I()
    elif name == 'L':
        return L()
    elif name == 'O':
        return O()
    elif name == 'P':
        return P()
    elif name =='S':
        return S()
    elif name == '/':
        return Slash()
    elif name == 'SP':
        return SP()
    elif name == 'SS':
        return SS()
    elif name == 'T':
        return T()
    elif name == 'TL':
        return TL()
    elif name == 'TR':
        return TR()
    elif name == 'X':
        return X()
    elif name == 'Z':
        return Z()
    else:
        raise InvalidFormat('Expected an edit descriptor, got %s' % name)

# All the tokens defined in the F77 specification unless specified

class A(object):
    def __init__(self):
        self.repeat = None
        self.width = None
    def __repr__(self):
        return '<A repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + '>'

class QuotedString(object):
    def __init__(self, char_string=None):
        self.char_string = char_string
    def get_width(self):
        return len(self.char_string)
    width = property(get_width)
    def __repr__(self):
        return '<QuotedString char_string=' + str(self.char_string) + '>'

# Only in F95
class B(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<B repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'
    
class BN(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<BN>'

class BZ(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<BZ>'

class Colon(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<Colon>'
    
class D(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
    def __repr__(self):
        return '<D repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + '>'

class E(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<E repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'
    
# Only in F95
class EN(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<EN repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'

# Only in F95
class ES(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<ES repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'

class F(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
    def __repr__(self):
        return '<F repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + '>'
    
class FormatGroup(object):
    pass

class G(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<G repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'

# Only in F77
class H(object):
    def __init__(self):
        self.num_chars = None
        self.char_string = None
    def __repr__(self):
        return '<H num_chars=' + str(self.num_chars) + \
                ' char_string=' + str(self.char_string) + '>'
    
class I(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<I repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'
    
class L(object):
    def __init__(self):
        self.repeat = None
        self.width = None
    def __repr__(self):
        return '<L repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + '>'

# Only in F95
class O(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<O repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'

class P(object):
    def __init__(self):
        self.scale = None
    def __repr__(self):
        return '<P scale=' + str(self.scale) + '>'
    
class S(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<S>'
    
class Slash(object):
    def __init__(self):
        self.repeat = None
        pass
    def __repr__(self):
        return '<Slash repeat=' + str(self.repeat) + '>'
    
class SP(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<SP>'
    
class SS(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<SS>'
    
class T(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<T num_chars=' + str(self.num_chars) + '>'
    
class TL(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<TL num_chars=' + str(self.num_chars) + '>'
    
class TR(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<TR num_chars=' + str(self.num_chars) + '>'

class X(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<X num_chars=' + str(self.num_chars) + '>'

# Only in F95
class Z(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<Z repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'

# Categorise the edit descriptors depnding on how they should be parsed

ED1 = ['BN', 'BZ', 'SP', 'SS', 'S'] # Of form X only
ED2 = ['X'] # Of form nX only
ED3 = ['T', 'TR', 'TL', 'L'] # Of form Xn only
ED4 = ['A'] # Of form X or Xn
ED5 = ['D', 'F'] # Of form Xn.m only
ED6 = ['B', 'I', 'O', 'Z'] # Of form Xn or Xn.m
ED7 = ['E', 'EN', 'ES', 'G'] # Of form Xn.m or Xn.mEe
ED8 = ['P'] # Of form kX only, where k is a signed integer, may omit comma if followed by Type 5 or 7 edit descriptor
ED9 = [':'] # Of form X only, may omit comma either side
ED10 = ['/'] # Of form X only, may omit following comma and leading comma if no repeat
REPEATABLE_EDS = ['L', 'A', 'D', 'F', 'B', 'I', 'O', 'Z', 'E', 'EN', 'ES', 'G', '/']
OUTPUT_EDS = (L, A, D, F, B, I, O, Z, E, EN, ES, G)
CONTROL_EDS = (BN, BZ, P, SP, SS, S, X, T, TR, TL, Colon, Slash)
NON_REVERSION_EDS = (P, S, SP, SS, BN, BZ)
ALL_ED = ED1 + ED2 + ED3 + ED4 + ED5 + ED6 + ED7 + ED8 + ED9 + ED10


    
'''
Miscellaneous functions, classes etc
'''

class has_next_iterator(object):
  '''
  A wrapper class for iterators so that the .has_next() method is implemented

  See - http://stackoverflow.com/questions/1966591/hasnext-in-python-iterators
  '''
  def __init__(self, it):
    self.it = iter(it)
    self._has_next = None
  def __iter__(self): return self
  def __next__(self):
    if self._has_next:
      result = self._the_next
    else:
      if IS_PYTHON3:
        result = next(self.it)
      else:
        result = self.it.next()
    self._has_next = None
    return result
  def next(self):
    if self._has_next:
      result = self._the_next
    else:
      if IS_PYTHON3:
        result = next(self.it)
      else:
        result = self.it.next()
    self._has_next = None
    return result
  def has_next(self):
    if self._has_next is None:
      try: 
        if IS_PYTHON3:
          self._the_next = next(self.it)
        else:
          self._the_next = self.it.next()
      except StopIteration: self._has_next = False
      else: self._has_next = True
    return self._has_next


def expand_edit_descriptors(eds):
    expanded_eds = []
    for ed in eds:
        if hasattr(ed, 'repeat') and (ed.repeat is not None):
            expanded_eds.extend(ed.repeat * [ed])
        else:
            expanded_eds.append(ed)
    return expanded_eds


# Should all edit descriptor values be returned even if they were not
# written to?
RET_WRITTEN_VARS_ONLY = False
# Should 'None' values be returned when no record is available to read
# from or the FORTRAN 'default'?
RET_UNWRITTEN_VARS_NONE = True
# The order in which edit desciptors are tried by default when G edit
# descriptor encountered on input
G_INPUT_TRIAL_EDS = ['F', 'L', 'A']
# Contrary to specification, many compilers allow zero width edit
# descriptors
ALLOW_ZERO_WIDTH_EDS = True
# Set the characters that separate the records
RECORD_SEPARATOR = '\n'

# The maximum size for an integer
if sys.version_info[0] >= 3:
    PROC_MAXINT = sys.maxsize
else:
    PROC_MAXINT = sys.maxint
# Processor dependant default for including leading plus or not
PROC_INCL_PLUS = False 
# Option to allow signed binary, octal and hex on input (not a FORTRAN feature)
PROC_ALLOW_NEG_BOZ = False
# Prcessor dependant padding character
PROC_PAD_CHAR = ' '
# Interpret blanks or jsut a negative as a zero, as in ifort behaviour
PROC_NEG_AS_ZERO = True
# Show a sign for zero?
PROC_SIGN_ZERO = False
PROC_MIN_FIELD_WIDTH = 46
PROC_DECIMAL_CHAR = '.'
G0_NO_BLANKS = False
PROC_NO_LEADING_BLANK = False
# The default value if BN, BZ edit descriptors are not specified
PROC_BLANKS_AS_ZEROS = False

def reset():
    global RET_WRITTEN_VARS_ONLY, RET_UNWRITTEN_VARS_NONE, PROC_INCL_PLUS, \
        PROC_ALLOW_NEG_BOZ, PROC_PAD_CHAR, PROC_NEG_AS_ZERO, PROC_SIGN_ZERO, \
        PROC_MIN_FIELD_WIDTH, PROC_DECIMAL_CHAR, G0_NO_BLANKS, \
        PROC_NO_LEADING_BLANK, PROC_BLANKS_AS_ZEROS, PROC_MAXINT, G_INPUT_TRIAL_EDS, \
        ALLOW_ZERO_WIDTH_EDS
    G_INPUT_TRIAL_EDS = ['F', 'L', 'A']
    if sys.version_info[0] >= 3:
        PROC_MAXINT = sys.maxsize
    else:
        PROC_MAXINT = sys.maxint
    RET_WRITTEN_VARS_ONLY = False
    RET_UNWRITTEN_VARS_NONE = True
    PROC_INCL_PLUS = False
    PROC_ALLOW_NEG_BOZ = False
    PROC_PAD_CHAR = ' '
    PROC_NEG_AS_ZERO = True
    PROC_SIGN_ZERO = False
    PROC_MIN_FIELD_WIDTH = 46
    PROC_DECIMAL_CHAR = '.'
    G0_NO_BLANKS = False
    PROC_NO_LEADING_BLANK = False
    PROC_BLANKS_AS_ZEROS = False
    ALLOW_ZERO_WIDTH_EDS = True

    
WIDTH_OPTIONAL_EDS = [A]
NON_WIDTH_EDS = [BN, BZ, P, SP, SS, S, X, T, TR, TL, Colon, Slash]
FORBIDDEN_EDS = [QuotedString, H]

# Some problems without pre written input vars:
#   Cannot say when reversion conditions are met
#   Cannot determine width of A edit descriptor
#   Cannot determine complex input
#   Cannot determine proper input for G edit descriptors


def _input(eds, reversion_eds, records, num_vals=None):

    state = { \
        'position' : 0,
        'scale' : 0,
        'incl_plus' : False,
        'blanks_as_zeros' : PROC_BLANKS_AS_ZEROS,
        # TODO: Implement halt if no more record input
        'halt_if_no_vals' : False,
        'exception_on_fail' : True,
    }

    # pdb.set_trace()

    for ed in eds + reversion_eds:
        if isinstance(ed, tuple(FORBIDDEN_EDS)):
            raise InvalidFormat("%d edit descriptr not permitted on input")

    # Expand repeated edit decriptors
    eds = expand_edit_descriptors(eds)
    reversion_eds = expand_edit_descriptors(reversion_eds)
    # Assume one-to-one correspondance between edit descriptors and output
    # values if number of output values is not defined 
    num_out_eds = 0
    for ed in eds:
        if isinstance(ed, OUTPUT_EDS):
            num_out_eds += 1
    num_rev_out_eds = 0
    if num_vals is None:
        num_vals = num_out_eds
    for ed in reversion_eds:
        if isinstance(ed, OUTPUT_EDS):
            num_rev_out_eds += 1

    
    # Will loop forever is no output edit descriptors
    if (num_out_eds == 0):
        return []
    # Will loop forever if no output eds in reversion format and is more values
    # requested than in the format
    if (num_vals > num_out_eds) and (num_rev_out_eds == 0):
        raise ValueError('Not enough output edit descriptors in reversion format to output %d values' % num_vals)

    # May need to process multiple records, down to a higher function to supply
    # appropriate string for format
    if not hasattr(records, 'next'):
        records = iter(re.split('\r\n|\r|\n', records))
    record = _next(records, None)
    if record is None:
        return [] 
    
    # if a_widths is not None:
    #     a_widths = itertools.cycle(a_widths)

    vals = []
    finish_up = False
    ed_ind = -1
    while True:
        ed_ind += 1
        # Signal to stop when Colon edit descriptor or end of format or end of
        # reversion format reached. Also not to output any more data
        if len(vals) >= num_vals:
            finish_up = True
        # Select the appropriate edit descriptor
        if ed_ind < len(eds):
            ed = eds[ed_ind]
        else:
            rev_ed_ind = (ed_ind - len(eds)) % len(reversion_eds)
            # Reversion begun and has been instructed to halt
            if finish_up and (rev_ed_ind == 0):
                break
            ed = reversion_eds[rev_ed_ind]

        if isinstance(ed, QuotedString):
            raise InvalidFormat('Cannot have string literal in an input format')
        elif isinstance(ed, BN):
            state['blanks_as_zeros'] = False
        elif isinstance(ed, BZ):
            state['blanks_as_zeros'] = True
        elif isinstance(ed, P):
            state['scale'] = ed.scale
        elif isinstance(ed, SP):
            state['incl_plus'] = True
        elif isinstance(ed, SS):
            state['incl_plus'] = False
        elif isinstance(ed, S):
            state['incl_plus'] = PROC_INCL_PLUS
        elif isinstance(ed, (X, TR)):
            state['position'] = min(state['position'] + ed.num_chars, len(record))
        elif isinstance(ed, TL):
            state['position'] = max(state['position'] - ed.num_chars, 0)
        elif isinstance(ed, T):
            if (ed.num_chars - 1) < 0:
                state['position'] = 0
            elif ed.num_chars > len(record):
                state['position'] = len(record)
            else:
                state['position'] = ed.num_chars - 1
        elif isinstance(ed, Slash):
            # End of record
            record = _next(records, None)
            state['position'] = 0
            if record is None:
                break
        elif isinstance(ed, Colon):
            # Break if input value satisfied
            if finish_up:
                break
        elif isinstance(ed, (Z, O, B, I)):
            val, state = read_integer(ed, state, record)
            vals.append(val)
        elif isinstance(ed, A):
            val, state = read_string(ed, state, record)
            vals.append(val)
        elif isinstance(ed, L):
            val, state = read_logical(ed, state, record)
            vals.append(val)
        elif isinstance(ed, (F, E, D, EN, ES)):
            val, state = read_float(ed, state, record)
            vals.append(val)
        elif isinstance(ed, G):
            # Difficult to know what wanted since do not know type of input variable
            # Use the G_INPUT_TRIAL_EDS variable to try the variables
            # until one sticks
            # n.b. vals and state do not get written to if
            # exception id raised
            resolved = False
            g_trial_eds = iter(G_INPUT_TRIAL_EDS)
            while not resolved:
                ed_name = _next(g_trial_eds, '')
                if ed_name.upper() in ('F', 'E', 'D', 'EN', 'ES'):
                    trial_ed = F()
                    trial_ed.width = ed.width
                    trial_ed.decimal_places = ed.decimal_places
                    # pdb.set_trace()
                    try:
                        val, state = read_float(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name.upper() in ('Z', 'O', 'B', 'I'):
                    trial_ed = globals()[ed_name]()
                    trial_ed.width = ed.width
                    trial_ed.min_digits = ed.decimal_places
                    try:
                        val, state = read_integer(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name.upper() in ('L'):
                    trial_ed = L()
                    trial_ed.width = ed.width
                    try:
                        val, state = read_logical(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name.upper() in ('A'):
                    trial_ed = A()
                    trial_ed.width = ed.width
                    try:
                        val, state = read_string(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name in ('G'):
                    raise ValueError('G edit descriptor not permitted in G_INPUT_TRIAL_EDS')
                else:
                    raise ValueError('Unrecognised trial edit descriptor string in G_INPUT_TRIAL_EDS')
                    
    if RET_WRITTEN_VARS_ONLY:
        vals = [val for val in vals if val is not None]
    return vals[:num_vals]

def _interpret_blanks(substr, state):
    # Save leading blanks
    len_str = len(substr)
    if state['blanks_as_zeros']:
        # TODO: Are tabs blank characters?
        substr = substr.replace(' ', '0')
    else:
        substr = substr.replace(' ', '')
    # If were blanks but have been stripped away, replace with a zero
    if len(substr) == 0 and (len_str > 0):
        substr = '0'
    return substr

def _get_substr(w, record, state):
    start = max(state['position'], 0)
    end = start + w
    # if end > len(record):
    #     substr = ''
    #     # TODO: test if no chars transmitted, then poition does not change
    #     w = 0
    # else:
    substr = record[start:end]
    state['position'] = min(state['position'] + w, len(record))
    return substr, state


def _next(it, default=None):
    try:
        if IS_PYTHON3:
            val = next(it)
        else:
            val = it.next()
    except StopIteration:
        val = default
    return val


def read_string(ed, state, record):
    if ed.width is None:
        # Will assume rest of record is fair game for the
        # unsized A edit descriptor
        ed.width = len(record) - state['position']
    substr, state = _get_substr(ed.width, record, state)
    val = substr.ljust(ed.width, PROC_PAD_CHAR)
    return (val, state)


def read_integer(ed, state, record):
    substr, state = _get_substr(ed.width, record, state)
    if ('-' in substr) and (not PROC_ALLOW_NEG_BOZ) and isinstance(ed, (Z, O, B)):
        if state['exception_on_fail']:
            raise ValueError('Negative numbers not permitted for binary, octal or hex')
        else:
            return (None, state)
    if isinstance(ed, Z):
        base = 16
    elif isinstance(ed, I):
        base = 10
    elif isinstance(ed, O):
        base = 8
    elif isinstance(ed, B):
        base = 2
    # If a negative is followed by blanks, Gfortran and ifort
    # interpret as a zero
    if re.match(r'^ *- +$', substr):
        substr = '0'
    # If a negative or negative and blanks, ifort interprets as
    # zero for an I edit descriptor
    if PROC_NEG_AS_ZERO and isinstance(ed, I) and re.match(r'^( *- *| +)$', substr):
        substr = '0'
    # If string is zero length (reading off end of record?),
    # interpret as zero so as to match what would be found in an
    # unwritten FORTRAN variable
    if substr == '':
        if RET_UNWRITTEN_VARS_NONE or RET_WRITTEN_VARS_ONLY:
            return (None, state)
        else:
            substr = '0'
    teststr = _interpret_blanks(substr, state)
    try:
        val = int(teststr, base)
    except ValueError:
        if state['exception_on_fail']:
            raise ValueError('%s is not a valid input for one of integer, octal, hex or binary' % substr)
        else:
            return (None, state)
    return (val, state)


def read_logical(ed, state, record):
    substr, state = _get_substr(ed.width, record, state)
    # Deal with case where there is no more input to read from
    if (substr == '') and (RET_UNWRITTEN_VARS_NONE or RET_WRITTEN_VARS_ONLY):
        return (None, state)
    # Remove preceding whitespace and take the first two letters as
    # uppercase for testing
    teststr = substr.upper().lstrip().lstrip('.')
    if len(teststr):
        teststr = teststr[0]
    else:
        # This is case where just a preceding period is read in
        raise ValueError('%s is not a valid boolean input' % substr)
    if teststr == 'T':
        val = True
    elif teststr == 'F':
        val = False
    else:
        if state['exception_on_fail']:
            raise ValueError('%s is not a valid boolean input' % substr)
        else:
            val = None
    return (val, state)


def read_float(ed, state, record):
    substr, state = _get_substr(ed.width, record, state)
    teststr = _interpret_blanks(substr, state)
    # When reading off end of record, get empty string,
    # interpret as 0
    if teststr == '':
        if RET_UNWRITTEN_VARS_NONE or RET_WRITTEN_VARS_ONLY:
            return (None, state)
        else:
            teststr = '0'
    # Python only understands 'E' as an exponential letter
    teststr = teststr.upper().replace('D', 'E')
    # Prepend an exponential letter if only a '-' or '+' denotes an exponent
    if 'E' not in teststr:
        teststr = teststr[0] + teststr[1:].replace('+', 'E+').replace('-', 'E-')
    # ifort allows '.' to be interpreted as 0
    if re.match(r'^ *\. *$', teststr):
        teststr = '0'
    # ifort allows '-' to be interpreted as 0
    if re.match(r'^ *- *$', teststr):
        teststr = '0'
    # ifort allows numbers to end with 'E', 'E+', 'E-' and 'D'
    # equivalents
    res = re.match(r'(.*)(E|E\+|E\-)$', teststr)
    if res:
        teststr = res.group(1)
    try:
        val = float(teststr)
    except ValueError:
        if state['exception_on_fail']:
            raise ValueError('%s is not a valid input as for an E, ES, EN or D edit descriptor' % substr)
        else:
            return (None, state)
    # Special cases: insert a decimal if none specified
    if ('.' not in teststr) and (ed.decimal_places is not None):
        val = val / 10 ** ed.decimal_places
    # Apply scale factor if exponent not supplied
    if 'E' not in teststr:
        val = val / 10 ** state['scale'] 
    return (val, state)

class FortranRecordReader(object):
    '''
    Generate a reader object for FORTRAN format strings

    Typical use case ...

    >>> header_line = FortranRecordReader('(A15, A15, A15)')
    >>> header_line.read('              x              y              z')
    ['              x', '              y', '              z']
    >>> line = FortranRecordReader('(3F15.3)')
    >>> line.read('          1.000          0.000          0.500')
    [1.0, 0.0, 0.5]
    >>> line.read('          1.100          0.100          0.600')
    [1.1, 0.1, 0.6]

    Note: it is best to create a new object for each format, changing the format
    causes the parser to reevalute the format string which is costly in terms of
    performance
    '''
    
    def __init__(self, format):
        self.format = format
        self._eds = []
        self._rev_eds = []
        self._parse_format()

    def __eq__(self, other):
        if isinstance(other, FortranRecordReader):
            return self.format == other.format
        else:
            return object.__eq__(self, other)

    def match(self, record):
        try:
            self.read(record)
        except RecordError:
            return False
        else:
            return True

    def read(self, record):
        '''
        Pass a string representing a FORTRAN record to obtain the relevent
        values
        '''
        return _input(self._eds, self._rev_eds, record)

    def get_format(self):
        return self._format
    def set_format(self, format):
        self._format = format
        self._parse_format()
    format = property(get_format, set_format)

    def _parse_format(self):
        self._eds, self._rev_eds = _parser(_lexer(self.format))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#/opt/local/bin/python
# Compute stellar Zanstra temperature and stellar luminosity.
# Last update: 2013-Oct-30 by RAS

import numpy as np
import scipy
import scipy.optimize as op
from scipy import constants as phy
from scipy.integrate import quad

WAV_H_IONIZ   = 911.7634
WAV_HE2_IONIZ = WAV_H_IONIZ/4.
ABMAG0     = 48.60
STMAG0     = 21.10
#FLUX_MAG0 = 3.68e-9    # from Pottasch (1984), outdated
FLUX_MAG0  = 3.63e-9    # or pow(10,-STMAG0/2.5)
MBOL_Sun   = 4.75

'''Determine the stellar Zanstra temperature from a nebular H or He emission 
   line flux and the stellar flux density. The method follows 
       Pottasch, S. 1984, "Planetary Nebulae" (Dordrecht: D. Reidel)
   with some modern updates. Also determine Bolometric correction and stellar 
   luminosity given temperature and distance. 

   Limitations: 
    - assumes Blackbody stellar SED
    - no correction for interstellar extinction.
'''

def logL(V, D_pc, T_eff):
    '''Calculate the stellar luminosity in Solar units, given an (unreddened) V 
       apparent magnitude, a distance (pc), and the stellar temperature (K).
    '''
    return -0.4*(V - 2.5*np.log10(D_pc) + BC(T_eff) - MBOL_Sun)


def BC(T_eff):
    '''Bolometric correction for very hot stars, from Vacca, Garmany & Shull,
       1996, ApJ, 460, 914.
    '''
    return 27.66 - 6.84*np.log10(T_eff)
#    return 28.46 - 7.08*np.log10(T_eff) + 0.08*log_g  # Fit incl. gravity

def integrand(x):
    return x**2/(np.exp(x)-1)

def G(ion,T):
    '''Compute integral over frequency for Planck function.
       ion - Reference Ion ("H | HE')
       T   - Stellar temperature (K)
       See Pottasch (1984), p. 169, eq. VII-8 for details. 
    '''

    if ion.upper()=='H':
        nu = phy.c / (WAV_H_IONIZ*1.e-10)
    elif ion.upper()=='HE2':
        nu = phy.c / (WAV_HE2_IONIZ*1.e-10)
    else:
        print('Unrecognized ion: %s' % ion)
        return None

    return quad(integrand, phy.h*nu/(phy.k*T), np.Inf)[0]


class Zanstra():
    def __init__(self, ion, logFlux=-10., magStar=10.):
        s = ion.upper()
        if s in ('H','HE2'):
            self.ion = s
        else:
            self.ion = 'H'
        self.logFlux = logFlux
        self.magStar = magStar

        if self.ion == 'H':
            self.ionScale = 3.95e-11
        elif self.ion == 'HE2':
            self.ionScale = 8.49e-11


    def z_ratio(self, T):
        '''Compute the ratio of the flux in a nebular emission line to the stellar 
           flux density for a given stellar (black-body) temperature T.
        '''

        return self.ionScale * T**3 * G(self.ion, T) * (np.exp(2.665e+4/T) - 1)


    def Func(self, T):
        '''Return the discriminant of the stellar Zanstra temperature function.
            '''

        r_obs = 10**self.logFlux / (FLUX_MAG0 * 10**(-self.magStar/2.5))
        return r_obs - self.z_ratio(T)


    def Solve(self):
        '''Determine the stellar Zanstra temperature from a nebular H or He 
           emission line flux and the stellar flux density.
        '''

        guess = 5.e+4
        return op.broyden1(self.Func, guess)



import pyneb as pn

def getTemDen_helper(inQ, outQ, atom, lev_i1, lev_j1, lev_i2, lev_j2,
         wave1, wave2, maxError, method, log, start_x, end_x, to_eval, nCut, maxIter,NLevels=None):
    ''' multiprocessing helper for getTemDen'''
    thisAtom = pn.Atom(atom=atom,NLevels=NLevels)
    for inputs in iter(inQ.get, 'STOP'):
        jobid = inputs[0]
        int_ratio = inputs[1]
        tem = inputs[2]
        den = inputs[3]
        res = thisAtom._getTemDen_1(int_ratio, tem=tem, den=den, lev_i1=lev_i1, lev_j1=lev_j1, lev_i2=lev_i2, lev_j2=lev_j2,
                  wave1=wave1, wave2=wave2, maxError=maxError, method=method, log=log, start_x=start_x, 
                  end_x=end_x, to_eval=to_eval, nCut=nCut, maxIter=maxIter)
        outQ.put((jobid, res))
    return



import pickle
import sys
import numpy as np
from scipy.interpolate import interp1d
from scipy import optimize

import pyneb as pn
from .pynebcore import RecAtom
from ..utils.physics import CST
from ..utils.misc import ROOT_DIR



class Continuum(object):
    
    def __init__(self):
        """
        Part of the PyNeb library.
        Mainly based on pySSN library        
        Adapted by V. Gomez-Llanos and C. Morisset, 2018
        """
        self.BE = None
        self.__HI_case = None
        self._set_HI_case('B')
        
    def _set_HI_case(self, case='B'):
        """
        Define the case (A or B) to be used for HI normalization line.
        Not sure that the Free-bound coefficients from Ercolano & Storey 2006 take case A option into account.
        """            
        if case != self.__HI_case:
            if case == 'A':
                pn.atomicData.setDataFile('h_i_rec_SH95-caseA.hdf5')
                self.HI = None
            elif case == 'B':
                pn.atomicData.setDataFile('h_i_rec_SH95.hdf5')
                self.HI = None
            else:
                raise ValueError('Unkown case {}. Should be A or B'.format(case))
            self.__HI_case = case
        
    def make_cont_Ercolano(self, tem, case, wl):
        """
        Adapted from http://adsabs.harvard.edu/abs/2006MNRAS.372.1875E
        tem : electron temperature [K]. Can not be an array. In case of tem array, use get_continuum
        case: one of "H", "He1", "He2"
        wl: wavelength [Angstrom]. May be a float or a numpy array
        return the continuum [erg/s.cm3/A]
        """
        try:
            _ = (e for e in wl)
        except TypeError:
            wl = np.array([wl])

        n_wl = len(wl)
        hnu =  CST.CLIGHT * 1e8 / wl * CST.HPLANCK  #!phy.c_ang_sec/wl*!phy.h
        if case == 'H':
            tab_T = 10**np.array([2. , 2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3. , 3.1, 3.2,
                                  3.3, 3.4, 3.5, 3.6, 3.7, 3.8, 3.9, 4. , 4.1, 4.2, 4.3, 4.4, 4.5,
                                  4.6, 4.7, 4.8, 4.9, 5. ])
            D = np.loadtxt(ROOT_DIR + '/' + 'atomic_data/coeff_ercolano_H.txt')
        elif case == 'He1':
            tab_T = 10**np.array([2.  , 2.04, 2.08, 2.12, 2.16, 2.2 , 2.24, 2.28, 2.32, 2.36, 2.4 ,
                                  2.44, 2.48, 2.52, 2.56, 2.6 , 2.64, 2.68, 2.72, 2.76, 2.8 , 2.84,
                                  2.88, 2.92, 2.96, 3.  , 3.04, 3.08, 3.12, 3.16, 3.2 , 3.24, 3.28,
                                  3.32, 3.36, 3.4 , 3.44, 3.48, 3.52, 3.56, 3.6 , 3.64, 3.68, 3.72,
                                  3.76, 3.8 , 3.84, 3.88, 3.92, 3.96, 4.  , 4.04, 4.08, 4.12, 4.16,
                                  4.2 , 4.24, 4.28, 4.32, 4.36, 4.4 , 4.44, 4.48, 4.52, 4.56, 4.6 ,
                                  4.64, 4.68, 4.72, 4.76, 4.8 , 4.84, 4.88, 4.92, 4.96, 5.  ])
            D = np.loadtxt(ROOT_DIR + '/' + 'atomic_data/coeff_ercolano_He1.txt')
        elif case == 'He2':
            tab_T = 10**np.array([2. , 2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3. , 3.1, 3.2,
                                  3.3, 3.4, 3.5, 3.6, 3.7, 3.8, 3.9, 4. , 4.1, 4.2, 4.3, 4.4, 4.5,
                                  4.6, 4.7, 4.8, 4.9, 5. ])
            D = np.loadtxt(ROOT_DIR + '/' + 'atomic_data/coeff_ercolano_He2.txt')
        else:
            print('Invalid case {0}'.format(case))
            return None
        if (tem < np.min(tab_T)) or (tem > np.max(tab_T)):
            print('Invalid temperature {0}'.format(tem))
            return None
        
        BE_E_Ry = D[:,1]
        BE_E_erg = BE_E_Ry * CST.RYD_erg
        BE_E_Thr = BE_E_erg[D[:,0] == 1]
        Delta_E = np.zeros(n_wl)
        for i in np.arange(n_wl):
            DE = hnu[i] - BE_E_Thr
            Delta_E[i] = np.min(DE[DE > 0])
            
        n_T_sup = np.min(np.where(tab_T >= tem)[0])
        n_T_inf = n_T_sup - 1
        T_sup = tab_T[n_T_sup]
        T_inf = tab_T[n_T_inf]
        
        BE_coeff_sup = D[:, n_T_sup+2]
        BE_coeff_inf = D[:, n_T_inf+2]
        
        coeff_sup = interp1d(BE_E_erg, BE_coeff_sup)(hnu)
        coeff_inf = interp1d(BE_E_erg, BE_coeff_inf)(hnu)

        C_interp= (np.log10(tem) - np.log10(T_inf)) / (np.log10(T_sup) - np.log10(T_inf))
        
        coeff = coeff_sup * C_interp + coeff_inf*(1. - C_interp)
        
        cont = coeff * 1e-34 * tem**(-1.5) * np.exp(-Delta_E / tem / CST.BOLTZMANN) / wl**2. * CST.CLIGHT * 1e8 # erg/s.cm3/A
        return cont.squeeze()

    def two_photon(self, tem, den, wl):
        """
        tem: temperature [K]. May be a float or a numpy array
        den: density [cm-3]
        wl: wavelength [Angstrom]. May be a float or a numpy array
        Return 2 photons continuum [erg/s.cm3/A]
        """
        #ToDo : See Schirmer, M. 2016, PASP, 128, 114001 
        try:
            _ = (e for e in wl)
        except TypeError:
            wl = np.array([wl])
        y = 1215.7 / wl
        A = 202.0 * (y * (1. - y) * (1. -(4. * y * (1 - y))**0.8) + 0.88 * ( y * (1 - y))**1.53 * (4. * y * (1 - y))**0.8)
        mask = y > 1.0 # Thanks to Daniel Schaerer for pointing out this potential issue
        A[mask] = 0.
        alfa_eff = 0.838e-13 * (tem / 1e4)**(-0.728) # fit DP de Osterbrock
        q = 5.31e-4 * (tem / 1e4)**(-0.17) # fit DP de Osterbrock
        n_crit = 8.226 / q
        twophot_cont = CST.HPLANCK * CST.CLIGHT * 1e8 / wl**3. * 1215.7 * A / 8.226 * alfa_eff / \
                       (1. + den/n_crit)                
        #twophot_cont[~np.isfinite(twophot_cont)] = 0.
        return twophot_cont
            
    def gff(self, Z, tem, wl):
        """
         Adaptated from http://adsabs.harvard.edu/abs/1991CoPhC..66..129S
        """
        D= np.array([8.986940175e+00, -4.009515855e+00,  8.808871266e-01,
            2.640245111e-02, -4.580645915e-02, -3.568055702e-03,   
            2.827798067e-03,  3.365860195e-04, -8.006936989e-01,
            9.466021705e-01,  9.043402532e-02, -9.608451450e-02,
            -1.885629865e-02,  1.050313890e-02,  2.800889961e-03, 
            -1.078209202e-03, -3.781305103e-01,  1.102726332e-01, 
            -1.543619180e-02,  8.310561114e-03,  2.179620525e-02, 
            4.259726289e-03, -4.181588794e-03, -1.770208330e-03,   
            1.877213132e-02, -1.004885705e-01, -5.483366378e-02,   
            -4.520154409e-03,  8.366530426e-03,  3.700273930e-03, 
            6.889320423e-04,  9.460313195e-05,  7.300158392e-02,   
            3.576785497e-03, -4.545307025e-03, -1.017965604e-02,   
            -9.530211924e-03, -3.450186162e-03,  1.040482914e-03, 
            1.407073544e-03, -1.744671550e-03,  2.864013856e-02,   
            1.903394837e-02,  7.091074494e-03, -9.668371391e-04,   
            -2.999107465e-03, -1.820642230e-03, -3.874082085e-04, 
            -1.707268366e-02, -4.694254776e-03,  1.311691517e-03, 
            5.316703136e-03,  5.178193095e-03,  2.451228935e-03,   
            -2.277321615e-05, -8.182359057e-04,  2.567331664e-04, 
            -9.155339970e-03, -6.997479192e-03, -3.571518641e-03, 
            -2.096101038e-04,  1.553822487e-03,  1.509584686e-03, 
            6.212627837e-04,  4.098322531e-03,  1.635218463e-03,   
            -5.918883504e-04, -2.333091048e-03, -2.484138313e-03, 
            -1.359996060e-03, -5.371426147e-05,  5.553549563e-04, 
            3.837562402e-05,  2.938325230e-03,  2.393747064e-03,   
            1.328839809e-03,  9.135013312e-05, -7.137252303e-04,   
            -7.656848158e-04, -3.504683798e-04, -8.491991820e-04, 
            -3.615327726e-04,  3.148015257e-04,  8.909207650e-04, 
            9.869737522e-04,  6.134671184e-04,  1.068883394e-04,   
            -2.046080100e-04 ])
    
        try:
            _ = (e for e in wl)
        except TypeError:
            wl = np.array([wl])
    
    
        XLF = np.log10(CST.CLIGHT * 1e8 / wl)
        N_wl = len(wl)
        G = np.zeros(N_wl)
        D = D.reshape(11, 8)
        B = np.zeros(11)
        C = np.zeros(8)
    
        XLRKT = 5.1983649 - np.log10(tem)
        TXG = 0.66666667 * (2.0 * np.log10(Z) + XLRKT)
    
        for j in np.arange(7):
            B[10] = D[10, j]
            B[9] = TXG * B[10] + D[9, j]
            for IR in np.arange(8)[::-1]:
                B[IR] = TXG * B[IR+1] - B[IR+2] + D[IR, j]
            C[j] = 0.25 * (B[0] - B[2])               
    
    
        CON=0.72727273 * XLRKT - 10.376127  
    
        for i in np.arange(N_wl):
            TXU = 0.72727273 * XLF[i] + CON 
            B[7] = C[7]
            B[6] = TXU * B[7] + C[6] 
            for IR in np.arange(5)[::-1]: 
                B[IR] = TXU * B[IR+1] - B[IR+2] + C[IR]
            G[i] = B[0] - B[2]
    
        return G.squeeze()

    def FreeFree(self, tem, wl, He1_H=0., He2_H=0., tem_HeI=None, tem_HeII=None):

        if tem_HeI is None:
            tem_HeI = tem
        if tem_HeII is None:
            tem_HeII = tem
            
        gff_HI = self.gff(1., tem, wl)
        gff_HeI = self.gff(1., tem_HeI, wl)
        gff_HeII = self.gff(4., tem_HeII, wl)

        FF_cont = (6.8391014e-38 * CST.CLIGHT * 1e8 / wl**2. * (
                        1.0**2. / np.sqrt(tem) * np.exp(-CST.HPLANCK*CST.CLIGHT*1e8/wl/CST.BOLTZMANN/tem) * gff_HI + 
                        He1_H * 1.0**2./ np.sqrt(tem_HeI) * np.exp(-CST.HPLANCK*CST.CLIGHT*1e8/wl/CST.BOLTZMANN/tem_HeI) * gff_HeI  + 
                        He2_H* 2.0**2. / np.sqrt(tem_HeII) * np.exp(-CST.HPLANCK*CST.CLIGHT*1e8/wl/CST.BOLTZMANN/tem_HeII) * gff_HeII))
        return FF_cont
    
    def _get_continuum1(self, tem, den, He1_H=0., He2_H=0., wl=None, 
                      cont_HI=True, cont_HeI=True, cont_HeII=True, 
                      cont_2p=True, cont_ff=True):
        """
        tem: temperature [K]. May be a float or an iterable
        den: density [cm-3]. May be a float or an iterable. If iterable, must have same size than tem
        He1_H and He2_H: He+/H+ and He++/H+ abundances. Default = 0.0
        wl: wavelengths: May be an array
        type of continuum to take into acount defined a boolean, defaults are True:
            cont_Hi, cont_HeI, cont_HeII: using B. Ercolano 2006 data
            cont_2p: 2 photons, using D. Pequignot fit to Osterbrock
        return the resulting continuum [erg/s.cm3/A]
        """

        cont = 0.
        if cont_HI:
            cont += self.make_cont_Ercolano(tem = tem, case = 'H', wl = wl)
        if cont_HeI and He1_H > 0.:
            cont += He1_H * self.make_cont_Ercolano(tem = tem, case = 'He1', wl = wl)
        if cont_HeII and He2_H > 0:
            cont += He2_H * self.make_cont_Ercolano(tem = tem, case = 'He2', wl = wl)
        if cont_2p:
            cont += self.two_photon(tem = tem, den = den, wl = wl)
        if cont_ff:
            cont += self.FreeFree(tem = tem, wl = wl)
        
        return cont

    def get_continuum(self, tem, den, He1_H=0., He2_H=0., wl=np.array([3500, 3600, 3700, 3800, 3900]), 
                      cont_HI=True, cont_HeI=True, cont_HeII=True, 
                      cont_2p=True, cont_ff=True, HI_label='11_2'):
        """
        tem: temperature [K]. May be a float or an iterable
        den: density [cm-3]. May be a float or an iterable. If iterable, must have same size than tem
        He1_H and He2_H: He+/H+ and He++/H+ abundances. Default = 0.0
        wl: wavelengths: May be an array
        type of continuum to take into acount defined a boolean, defaults are True:
            cont_Hi, cont_HeI, cont_HeII: using B. Ercolano 2006 data
            cont_2p: 2 photons, using D. Pequignot fit to Osterbrock
            cont_ff: Free-Free, from Storey & Hummer 1991
        HI_label: HI label to normalize the continuum. If None, no normalization is done. Default 11_2
        return the resulting continuum. Unit [A-1] if normalized, [erg/s.cm3/A] otherwise
        
        Exemple of use:
            C = pn.Continuum()
            wl = np.arange(3500, 4000, 1)
            cont = C.get_continuum(tem=1e4, den=1e2, He1_H=0.08, He2_H=0.02, wl=wl)
            plt.plot(wl, cont)
        """
        try:
            _ = (e for e in tem)
            T_iterable = True
            try:
                _ = (e for e in den)
            except:
                den = np.ones_like(tem) * den
        except TypeError:
            T_iterable = False
        if HI_label is None:
            norm = 1.0
        else:
            if self.HI is None:
                self.HI = pn.RecAtom('H',1)
            norm = self.HI.getEmissivity(tem, den, label = HI_label, product=False)
            
        if T_iterable:
            cont = np.array(list(map(lambda t, d: self._get_continuum1(t, d, He1_H=He1_H, He2_H=He2_H, wl=wl, 
                                                                       cont_HI=cont_HI, cont_HeI=cont_HeI, 
                                                                       cont_HeII=cont_HeII, 
                                                                       cont_2p=cont_2p, cont_ff=cont_ff), 
                                     tem, den))).T
            return cont.squeeze()/norm
        else:
            cont = self._get_continuum1(tem, den, He1_H=He1_H, He2_H=He2_H, wl=wl, 
                                        cont_HI=cont_HI, cont_HeI=cont_HeI, cont_HeII=cont_HeII, 
                                        cont_2p=cont_2p, cont_ff=cont_ff)
            return cont/norm
    
    def BJ_HI(self, tem, den, He1_H, He2_H, wl_bbj = 3643, wl_abj = 3861, HI_label='11_2'):
        """
        tem: temperature [K]. May be a float or an iterable
        den: density [cm-3]. May be a float or an iterable. If iterable, must have same size than tem
        He1_H and He2_H: He+/H+ and He++/H+ abundances.
        
        wl_bbj, wl_abj: wavelengths below and above the jump resp. Defaults are 3643 and 3861.
        HI_label: reference HI line to normalize the jump. Default is 11_2
        
        return the Balmer Jump (may be any other jump if wl are changed) normalized to the HI line
        """
        
        fl_bbj, fl_abj = self.get_continuum(tem = tem, den = den, He1_H = He1_H, 
                                            He2_H = He2_H, wl = np.array([wl_bbj, wl_abj]),
                                            HI_label=HI_label)
        
        BJ_HI = fl_bbj - fl_abj
        return BJ_HI
    
    def T_BJ(self, BJ_HI, den, He1_H, He2_H, wl_bbj = 3643, wl_abj = 3861, HI_label='11_2',
             T_min=5e2, T_max=3e4):
        """
        BJ_HI: Balmer Jump (may be any other jump if wl are changed) normalized to the HI line
        den: density [cm-3]. May be a float or an iterable. If iterable, must have same size than tem
        He1_H and He2_H: He+/H+ and He++/H+ abundances.
        
        wl_bbj, wl_abj: wavelengths below and above the jump resp. Defaults are 3643 and 3861.
        HI_label: reference HI line to normalize the jump. Default is 11_2
        
        T_min, T_max: limits for the root finding exploration
        
        return temperature [K] corresponding to the jump
        """
        try:
            _ = (e for e in BJ_HI)
            BJ_iterable = True
        except TypeError:
            BJ_iterable = False
            
        def f2minimize(tem, BJ_HI):
            f = self.BJ_HI(tem, den=den, He1_H=He1_H, He2_H=He2_H, wl_bbj = wl_bbj, wl_abj=wl_abj, HI_label=HI_label) - BJ_HI
            return f
            
        if BJ_iterable:
            T_BJ = np.array(list(map(lambda bjhi: optimize.brentq(f2minimize, T_min, T_max, args=bjhi), BJ_HI))).T
            return T_BJ.squeeze()
        else:
            T_BJ = optimize.brentq(f2minimize, T_min, T_max, args=BJ_HI)
            return T_BJ

    
    def T_BJ_Liu(self, BJ_H11, He1_H, He2_H):
        """
        From Liu, X.-W., Luo, S.-G., Barlow, M. J., Danziger, I. J., & Storey, P. J.
        2001, MNRAS, 327, 141-168

        """
        T = 368 * (1 + 0.259 * He1_H + 3.409 * He2_H) * (BJ_H11)**(-3./2)
        return T


import os
import numpy as np

from pyneb import config, log_, atomicData
if config.INSTALLED['plt']:
    import matplotlib.pyplot as plt
from ..utils.misc import int_to_roman, parseAtom2
from ..core.pynebcore import Atom
from ..utils.saverestore import save, restore

class EmisGrid(object):
    """
    Instantiates an atom and compute the emissivities of all the atom lines for the (tem, den) 
    values of a regularly spaced grid.
    Each line is represented in a 2D array, and there are as many arrays as transitions in the atom.
    The results can be saved for a later use in a cPickle file.

    """

    def __init__(self, elem=None, spec=None, n_tem=100, n_den=100, tem_min=5000., tem_max=20000.,
                 den_min=10., den_max=1.e8, restore_file=None, atomObj=None, NLevels=None, **kwargs):

        """
        Container for the emission line intensity grids depending on Te and Ne.

        Usage:
            O3map = pn.EmisGrid('O', 3, n_tem=30, n_den=30)
            O3map = pn.EmisGrid(restore_file='plot/O3_10k.pypic') # Recovers the map from a previous computation

        Parameters:
            - elem               element (e.g. 'O')
            - spec               spectrum (e.g. 3)
            - n_tem              number of points of the electron temperature sample
            - n_den              number of points of the electron density sample
            - tem_min, tem_max   limits for the electron temperature
            - den_min, den_max   limits for the electron density
            - restore_file       if set, the emissivities are loaded from this file, 
                                 otherwise (None) it is computed
                                 If the tem or den table from the restored file does not match the given parameters,
                                 a new grid is computed and self.compute_new_grid is set to True. The same applies
                                 if the atomic data are not the same.
            - atomObj            an atom object. In this case, no need for elem and spec.

        """
        self.log_ = log_
        self.calling = 'EmisGrid'

        if restore_file is not None:
            old = restore(restore_file)
            self.elem = old['elem']
            self.spec = old['spec']
            if atomObj is None:
                self.atom = Atom(elem=self.elem, spec=self.spec, NLevels=NLevels, **kwargs)
            else:
                self.atom = atomObj
            if self.atom.atomFitsFile != old['atomFitsFile']:
                self.log_.error('You are using {0}, but restoring a file made with {1}'.format(self.atom.atomFitsFile, old['atomFitsFile']),
                                calling=self.calling)
            if self.atom.collFitsFile != old['collFitsFile']:
                self.log_.error('You are using {0}, but restoring a file made with {1}'.format(self.atom.collFitsFile, old['collFitsFile']),
                                calling=self.calling)
            
            self.compute_new_grid = False
            if n_tem != len(old['tem']):
                self.log_.warn('len(tem) does not match saved data. New grid is computed')
                self.compute_new_grid = True
            else:
                if not np.isclose(tem_min, min(old['tem'])):
                    self.log_.warn('Min(tem) does not match saved data. New grid is computed')
                    self.compute_new_grid = True
                if not np.isclose(tem_max, max(old['tem'])):
                    self.log_.warn('Max(tem) does not match saved data. New grid is computed')
                    self.compute_new_grid = True
            if n_den != len(old['den']):
                self.log_.warn('len(den) does not match saved data. New grid is computed')
                self.compute_new_grid = True
            else:
                if not np.isclose(den_min, min(old['den'])):
                    self.log_.warn('Min(den) does not match saved data. New grid is computed')
                    self.compute_new_grid = True
                if not np.isclose(den_max, max(old['den'])):
                    self.log_.warn('Max(den) does not match saved data. New grid is computed')
                    self.compute_new_grid = True
            if not self.compute_new_grid:
                self.tem = old['tem']
                self.den = old['den']
                self.n_tem = self.tem.size
                self.n_den = self.den.size
                self.tem_min = min(self.tem)
                self.tem_max = max(self.tem)
                self.den_min = min(self.den)
                self.den_max = max(self.den)
                self.emis_grid = old['emis_grid']
        else:
            self.compute_new_grid = True
            
        if self.compute_new_grid:
            if atomObj is None:
                self.elem = elem
                self.spec = spec
                self.atom = Atom(elem, spec, NLevels=NLevels, **kwargs)
            else:
                self.atom = atomObj
                self.elem = self.atom.elem
                self.spec = self.atom.spec
            self.n_tem = n_tem
            self.n_den = n_den
            self.tem_min = tem_min
            self.tem_max = tem_max
            self.den_min = den_min
            self.den_max = den_max
            self.tem = 10 ** np.linspace(np.log10(tem_min), np.log10(tem_max), n_tem)
            self.den = 10 ** np.linspace(np.log10(den_min), np.log10(den_max), n_den)
            try:
                self.atomFitsFile = self.atom.atomFitsFile
                self.collFitsFile = self.atom.collFitsFile
            except:
                self.atomFitsFile = None
                self.collFitsFile = None
            try:
                self.recFitsFile = self.atom.recFitsFile
            except:
                self.recFitsFile = None
            self.emis_grid = self.atom.getEmissivity(self.tem, self.den)
            if restore_file is not None:
                self.save(restore_file)
                self.log_.message('%s saved to %s' % ((self.atom), restore_file), calling=self.calling)
            #self.emis_grid = np.empty((self.atom.NLevels, self.atom.NLevels, n_tem, n_den))
        
        self.den2D, self.tem2D = np.meshgrid(self.den, self.tem)
        
#    def computeEmis(self):
#        """
#        This is where the emissivities are computed, by calling a function from Atom.
#        """
#        self.emis_grid = self.atom.getEmissivity(self.tem, self.den)

    def save(self, file_):
        """
        Save results for a later use.
        
        Usage:
            O3map.save('plot/O3_30by30.pypic')
        
        Parameter:
            file_  the name of the file in which the emissivity grids are stored

        """
        save(file_, emis_grid=self.emis_grid, tem=self.tem, den=self.den, elem=self.elem,
             spec=self.spec, atomFitsFile=self.atomFitsFile, collFitsFile=self.collFitsFile)


    def getGrid(self, lev_i=None, lev_j=None, wave= -1, to_eval=None, label=None):
        """
        2D array of a line emissivity for the (Te, Ne) values of a regularly spaced grid.
        The line is specified either as the wavelength or the levels. An expression can also be used, 
        for example to_eval = 'L(5007)/L(4959)'
        
        Parameters:
            - lev_i, lev_j   levels for the emission line
            - wave           wavelength
            - to_eval        algebraic expression of the line combination to evaluate
        
        """
        if wave != -1:
            lev_i, lev_j = self.atom.getTransition(wave)
        if label is not None:
            to_eval = 'S("{}")'.format(label)

        if to_eval is None:
            to_eval = 'I(' + str(lev_i) + ',' + str(lev_j) + ')'
            
        def I(lev_i, lev_j):
            return self.emis_grid[lev_i - 1, lev_j - 1, :, :]
        def L(wave):
            return self.getGrid(wave=wave)
        def S(label):
            return self.emis_grid[label]
        return eval(to_eval)
    
    
    def plotImage(self, to_eval=None, lev_i1=None, lev_j1=None, lev_i2=None, lev_j2=None,
                  wave1= -1, wave2= -1, cblabel='', **kwargs):
        """
        Draw a contour plot of a line ratio. The contour is obtained by a horizontal cut in a temden-emissivity array.
        More elaborated plots can be obtained with the Diagnostic object.
        
        Usage:
            O3map.plotImage('L(4363)/L(5007)')
            O3map.plotImage(wave1=4363, wave2=5007)
            
        Parameters:
            - to_eval                         algebraic expression of the line combination to draw. May combine L(lambda) and I(i,j).
            - lev_i1, lev_i2, lev_j1, lev_j2  levels of the a lines in case of simple line ratio plot.
            - wave1, wave2                    wavelengths of the two lines in case of simple line ratio plot.
            - cblabel                         a title for the colorbar
            - **kwargs                        any other parameter will be sent to contourf and pcolor

        """
        if not config.INSTALLED['plt']:
            log_.error('Matplotlib not available', calling=self.calling)
            return None
        L = lambda lam: self.getGrid(wave=lam)
        I = lambda i, j: self.getGrid(i, j)
        S = lambda label: self.getGrid(label)
        if to_eval is None:
            if wave1 != -1:
                lev_i1, lev_j1 = self.atom.getTransition(wave1)
            if wave2 != -1:
                lev_i2, lev_j2 = self.atom.getTransition(wave2)
            to_eval = 'I(' + str(lev_i1) + ',' + str(lev_j1) + ') / I(' + str(lev_i2) + ',' + str(lev_j2) + ')'
        X = np.log10(self.den2D)
        Y = self.tem2D / 1e4
        Z = self.getGrid(to_eval=to_eval)
        plt.contourf(X, Y, Z, **kwargs)
        emap = plt.pcolor(X, Y, Z, **kwargs)
        cbar = plt.colorbar(emap)
        cbar.set_label(cblabel)
        # plt.clabel(contour, inline=1)
        plt.xlabel(r'log Electron density (cm$^{-3}$)')
        plt.ylabel(r'Electron temperature ($10^4$K)')
        plt.show()


    def plotContours(self, to_eval, low_level=None, high_level=None, n_levels=20,
                      linestyles='-', clabels=True, log_levels=True, title=None, ax=None, **kwargs):
        """
        Plot a contour map of the diagnostic defined by to_eval.
        
        Usage:
            O3map.plotContours('L(4363)/L(5007)')
        
        Parameters:
            - to_eval                   algebraic definition of the line ratio to contour-plot
            - low_levels, high_levels   limit levels for the contour. If not set, the limit of the 
                                        ratio values are used.
            - n_levels                  number of levels (20 is the default)
            - linestyles                used for the contour lines (default: solid line)
            - clabels                   Boolean. Controls if line labels are printed (default: True)
            - log_levels                Boolean. If True (default), the log of the line ratio is used.
            - title                     plot title
            - **kwargs                  sent to plt.contour
            
        """ 
        if not config.INSTALLED['plt']:
            log_.error('Matplotlib not available', calling=self.calling)
            return None
        X = np.log10(self.den2D)
        Y = self.tem2D
        L = lambda lam: self.getGrid(wave=lam)
        I = lambda i, j: self.getGrid(i, j)
        S = lambda label: self.getGrid(label=label)
        try:
            diag_map = eval(to_eval)
        except:
            self.log_.warn('diag %s not found' % to_eval, calling=self.calling)
            return None
        if low_level is None:
            low_level = np.min(np.log10(diag_map))
        if high_level is None:
            high_level = np.max(np.log10(diag_map))
        if log_levels:
            Z = np.log10(diag_map)
            levels = np.linspace(low_level, high_level, n_levels)
            if title is None:
                title = '[%s%s]: log(%s)' % (self.elem, int_to_roman(int(self.spec)), to_eval)
        else:
            Z = diag_map
            levels = 10. ** (np.linspace(low_level, high_level, n_levels))
            if title is None:
                title = '[%s%s]: %s' % (self.elem, int_to_roman(int(self.spec)), to_eval)
        
        if ax is None:
            f, ax = plt.subplots()
        else:
            f = plt.gcf()
        CS = ax.contour(X, Y, Z, levels=levels, linestyles=linestyles, **kwargs)
        if clabels:
            ax.clabel(CS, CS.levels[::3], inline=True, fontsize=12, colors='black')
        ax.set_xlabel(r'log(n$_{\rm e}$) [cm$^{-3}$]')
        ax.set_ylabel(r'T$_{\rm e}$ [K]')
        ax.set_title(title)


    def plotLineRatio(self, to_eval, par=None, par_low=None, par_high=None, n_par=10,
                      linestyles='-', title=None, legend=True, loc=2, **kwargs):
        """
        Plot the diagnostic defined by to_eval as a function of either Ne or Te, with the other
        variable as a parameter.
        
        Usage:
            o3grid.plotLineRatio('L(4363)/L(5007)', par='den')
            s2grid.plotLineRatio('I(2,1)/I(3,1)', par='tem', par_low=5000, par_high=20000, n_par=4)
        
        Parameters:
            - to_eval      algebraic definition of the line ratio to contour-plot
            - par          quantity to be used as a parameter (either 'den' or 'tem')
            - par_low      lowest limit of the parameter range
            - par_high     highest limit of the parameter range
            - n_par        number of parameter's values (default: 10)
            - linestyles   used for the contour lines (default: solid)
            - title        plot title
            - legend       Boolean. If True, write legend title (default: True) 
            - **kwargs     sent to plt.contour
            
        """ 
        if not config.INSTALLED['plt']:
            log_.error('Matplotlib not available', calling=self.calling)
            return None

        L = lambda wav: self.getGrid(wave=wav)
        I = lambda i, j: self.getGrid(i, j)
        
        if par == 'tem':
            X = np.log10(self.den2D)
            Z = self.tem2D
            leg_format = "{0:.0f} K"
            leg_title = r'Temperature'
            xlabel = r'Log(N$_{\rm e}$) [cm$^{-3}$]'
        elif par == 'den':
            X = self.tem2D
            Z = np.log10(self.den2D)
            leg_format = "{0:.2f}"
            leg_title = r'Log(N$_{\rm e}$)'
            xlabel = r'T$_{\rm e}$ [K]'
        else:
            self.log_.error('The parameter must be either den or tem (%s given)' % par, calling=self.calling)

        try:
            Y = eval(to_eval)
        except:
            self.log_.warn('Diagnostic %s not found' % to_eval, calling=self.calling)            
            return None

        if par_low is None:
            par_low = np.min(Z)
        if par_high is None:
            par_high = np.max(Z)
        levels = np.linspace(par_low, par_high, n_par)

        CS = plt.contour(X, Y, Z, levels=levels, linestyles=linestyles, **kwargs)
        
        if legend:
            for i_level in range(len(levels)):
                CS.collections[i_level].set_label(leg_format.format(levels[i_level]))
                plt.legend(title=leg_title, loc=loc)

        if title is None:
            title = '[%s%s]' % (self.elem, int_to_roman(int(self.spec)))

        plt.xlabel(xlabel)
        plt.ylabel(to_eval)
        plt.title(title)
    
        plt.show()
        

def getEmisGridDict(elem_list=None, spec_list=None, atom_list=None, restore=True, pypic_path=None,
            n_tem=100, n_den=100, tem_min=5000., tem_max=20000., den_min=10., den_max=1e8, save=True,
            atomDict=None, createPypicsDir=True, computeIfAbsent=True):
    """
    Return a dictionary of EmisGrid objects referred to by their atom string (e.g. 'O3').
    
    Usage:
        emisgrids = pn.getEmisGridDict(['C', 'N', 'O'], [2, 3], save=True)
        emisgrids = pn.getEmisGridDict(atom_list=['N2', 'O2', 'O3'], pypic_path='/tmp/pypics/',
                        den_max = 1e6)
        emisgrids = pn.getEmisGridDict(atomDict=diags.atomDict)

    Parameters:
        - elem_list         list of elements
        - spec_list         list of spectrum values (integers)
        - atom_list         list of atoms (e.g. ['N2', 'O3'])
        - restore           Boolean. If True (default), the program will search for a previously computed grid.
                            If not found, it will compute the EmisGrid (unless computeIfAbsent is set to False)
        - pypic_path        directory where to look for the pypic emissivity files. It defaults to pn.config.pypic_path, 
                            which is defined when the PyNeb session begins and corresponds to $HOME/.pypics if
                            the ENVIRONMENT variable HOME is accessible or to /tmp/pypics otherwise
        - n_tem, n_den      number of points in the Te and Ne samples resp. 
        - tem_min, tem_max, den_min, den_max    limits for Te and Ne
        - save              Boolean. If True (default), the emissivity grid will be saved. It has no effect if the 
                            EmisGrid has just been restored.
        - atomDict          dictionary of Atom objects
        - computeIfAbsent   If the file to restore is absent and this parameter is True (default), the 
                            EmisGrid is computed
            
    """
    if pypic_path is None:
        pypic_path = config.pypic_path
    else:
        config.pypic_path = pypic_path
    calling = 'getEmisGridDict'
    if pypic_path is None:
        return None
    emis_grids = {}
    atoms = []
    if atomDict is not None:
        atoms = atomDict.keys()
    else:
        if atom_list is not None:
            atoms = atom_list
        else:
# VL Removed 13 March 2015 because it made the code ignore the elem keyword             
#            atoms = atomicData.getAllAtoms()
            if (elem_list is None) and (spec_list is None):
                atoms = atomicData.getAllAtoms()
            else:
                for elem in elem_list:
                    for spec in spec_list:
                        atoms.append(elem + str(spec)) 
    for atom in atoms:
        elem, spec, rec = parseAtom2(atom)
        if atomDict is not None:
            atomObj = atomDict[atom]
        else:
            atomObj = None
        file_ = '{0}/emis_{1}{2}{3}.pypic'.format(pypic_path, elem, spec, rec)
        if restore:
            if os.path.exists(file_):
                try:
                    emis_grids[elem + spec] = EmisGrid(restore_file=file_, n_tem=n_tem, n_den=n_den, 
                                                       tem_min=tem_min, tem_max=tem_max,
                                                       den_min=den_min, den_max=den_max, atomObj=atomObj)
                    log_.message('Read %s' % file_, calling=calling)
                    compute_this = False
                except:
                    if computeIfAbsent:
                        log_.warn('Wrong emission map: {0}, creating it'.format(file_), calling=calling)
                        compute_this = True
                    else:
                        log_.error('Wrong emission map: {0}'.format(file_), calling=calling)
                        compute_this = False
            else:
                log_.warn('Emission map not found: {0}'.format(file_), calling=calling)
                if computeIfAbsent:
                    compute_this = True
        else:
            compute_this = True
        if compute_this:
            try:
                emis_grids[atom] = EmisGrid(elem, spec, n_tem=n_tem, n_den=n_den, tem_min=tem_min, tem_max=tem_max,
                                              den_min=den_min, den_max=den_max, atomObj=atomObj)
                if save:
                    emis_grids[atom].save(file_)
                    log_.message('%s saved to %s' % ((atom), file_), calling=calling)
            except:
                log_.warn('No %s EmisGrid' % (atom), calling=calling)
    return emis_grids
